from truthdiscovery.graphs.animation import (
    BaseAnimator,
    GifAnimator,
    JsonAnimator
)
from truthdiscovery.graphs.backends import BaseBackend, PngBackend, JsonBackend
from truthdiscovery.graphs.colours import (
    GraphColourScheme,
    NodeType,
    PlainColourScheme,
    ResultsGradientColourScheme
)
from truthdiscovery.graphs.draw import (
    GraphRenderer,
    MatrixDatasetGraphRenderer,
)
from truthdiscovery.graphs.entities import (
    Circle,
    Entity,
    Label,
    Line,
    Rectangle
)



import itertools

from bidict import bidict
import scipy.sparse


class IDMapping(bidict):
    """
    Bi-directional mapping from *labels* (of arbitrary type) to integer IDs
    """
    def get_id(self, label, insert=True):
        """
        :param label:  label to return ID for
        :param insert: if True, create a new ID for ``label`` if it is not
                       already present (default: True)
        :return: the ID for the ``label``
        :raises KeyError: if ``insert=False`` and ``label`` is not present
        """
        if label not in self and insert:
            self[label] = len(self)
        return self[label]


class Dataset:
    """
    An object to represent a dataset upon which truth discovery will be
    performed. A dataset consists *variables* and the values they take
    according to different *sources*.

    Input is given as an iterable of tuples of the form
    ``(source_label, var_label, value)``. Labels and values in these tuples
    may be of any type that can be used as keys in a dict.

    Optionally the dataset may include implication values between claims about
    the same variable (e.g. for TruthFinder).

    The implication between claims ``var = x`` and ``var = y`` is a value in
    [-1, 1] that describes how the confidence that ``var = x`` influences the
    confidence of ``var = y``.  A positive value indicates that if ``var = x``
    is true, then ``var = y`` is likely to be true. A negative value means that
    if ``var = x`` is true, then ``var = y`` is likely to be false (Yin et.
    al., 2008).
    """
    source_ids = None
    var_ids = None
    claim_ids = None
    val_hashes = None

    def __init__(self, triples, allow_multiple=False,
                 implication_function=None):
        """
        :param triples:        iterable of ``(source_label, var_label, value)``
                               as described above
        :param allow_multiple: if a source makes multiple claims for the same
                               variable, use the first value instead of raising
                               ValueError
        :param implication_function: (optional) function to compute implication
                                     values between claims (see above). This
                                     should take ``(var, val1, val2)`` as
                                     arguments and return an implication value
                                     in [-1, 1], or None
        """
        self.source_ids = IDMapping()  # Map source label to integer IDs
        self.var_ids = IDMapping()     # Variable labels to IDs
        self.val_hashes = IDMapping()  # Values to IDs (hashes)
        self.claim_ids = IDMapping()   # (var_id, val_hash) tuples to IDs

        # Keep track of (source, var) pairs to detect if a source makes more
        # than one claim for a single variable
        source_var_pairs = {}

        # Keep track of (source_id, claim_id) pairs to populate the
        # source-claims matrix. Source and claim IDs for each pair are stored
        # in the same position in *separate lists*, since this is the format
        # the sparse matrix constructor requires
        sc_rows = []
        sc_cols = []

        # Keep track of all claims IDs for each variable, to populate the
        # mutual exclusion matrix. The keys are variable IDs, and values are
        # sets of claim IDs
        mut_ex_claims = {}

        for source_label, var_label, val in triples:
            s_id = self.source_ids.get_id(source_label)
            var_id = self.var_ids.get_id(var_label)
            val_hash = self.val_hashes.get_id(val)

            if (s_id, var_id) not in source_var_pairs:
                source_var_pairs[(s_id, var_id)] = True
            elif allow_multiple:
                continue
            else:
                raise ValueError(
                    "Source '{}' claimed more than one value for variable '{}'"
                    .format(self.source_ids.inverse[s_id],
                            self.var_ids.inverse[var_id])
                )

            claim = (var_id, val_hash)
            claim_id = self.claim_ids.get_id(claim)
            sc_rows.append(s_id)
            sc_cols.append(claim_id)

            if var_id not in mut_ex_claims:
                mut_ex_claims[var_id] = set()
            mut_ex_claims[var_id].add(claim_id)

        self.num_sources = len(self.source_ids)
        self.num_variables = len(self.var_ids)
        self.num_claims = len(self.claim_ids)

        # Create source-claim matrix: entry (i, j) is 1 if source i makes claim
        # j, and 0 otherwise
        self.sc = scipy.sparse.csr_matrix(
            ([1] * len(sc_rows), (sc_rows, sc_cols)),
            shape=(self.num_sources, self.num_claims)
        )

        # Create mutual exclusion matrix: entry (i, j) is 1 if claims i and j
        # relate to the same variable (including when i=j) and 0 otherwise
        mut_ex_rows = []  # Construct in the same way as for sc
        mut_ex_cols = []
        for claim_ids in mut_ex_claims.values():
            for i, j in itertools.product(claim_ids, repeat=2):
                mut_ex_rows.append(i)
                mut_ex_cols.append(j)

        self.mut_ex = scipy.sparse.csr_matrix(
            ([1] * len(mut_ex_rows), (mut_ex_rows, mut_ex_cols)),
            shape=(self.num_claims, self.num_claims)
        )

        # Create implication matrix, for implications between claims
        imp_rows = []
        imp_cols = []
        imp_entries = []
        if implication_function is not None:
            # Iterate over non-zero entries in mut ex
            for j1, j2 in zip(*self.mut_ex.nonzero()):
                if j1 == j2:
                    continue
                # Note that claims j1 and j2 are for the same variable, since
                # mut ex is 1 at this point
                var_id, val1_hash = self.claim_ids.inverse[j1]
                _, val2_hash = self.claim_ids.inverse[j2]

                var = self.var_ids.inverse[var_id]
                val1 = self.val_hashes.inverse[val1_hash]
                val2 = self.val_hashes.inverse[val2_hash]
                imp_value = implication_function(var, val1, val2)

                if imp_value is not None:
                    if imp_value < -1 or imp_value > 1:
                        raise ValueError(
                            "Implication values must be in [-1, 1]"
                        )
                    imp_entries.append(imp_value)
                    imp_rows.append(j1)
                    imp_cols.append(j2)

        if imp_entries:
            self.imp = scipy.sparse.csr_matrix(
                (imp_entries, (imp_rows, imp_cols)),
                shape=(self.num_claims, self.num_claims)
            )
        else:
            self.imp = scipy.sparse.csr_matrix(
                (self.num_claims, self.num_claims)
            )

    def get_belief_dict(self, claim_beliefs):
        """
        Convert belief in claims to belief in (var, val) pairs.

        :param belief: numpy array of belief values for claims, ordered by
                       claim ID
        :return:       a dict of belief values for variables taking different
                       values, in the format required for :any:`Result`
        """
        var_beliefs = {}
        for claim_id, belief_score in enumerate(claim_beliefs):
            var_id, val_hash = self.claim_ids.inverse[claim_id]
            var_label = self.var_ids.inverse[var_id]
            val = self.val_hashes.inverse[val_hash]
            if var_label not in var_beliefs:
                var_beliefs[var_label] = {}
            var_beliefs[var_label][val] = belief_score
        return var_beliefs

    def get_source_trust_dict(self, trust):
        """
        :param trust: numpy array of source trust values, ordered by source ID
        :return:      a dict of source trusts in the format required for
                      :any:`Result`
        """
        return {
            self.source_ids.inverse[i]: trust_val
            for i, trust_val in enumerate(trust)
        }

    def num_connected_components(self):
        """
        :return: the number of connected components in the graph representation
                 of the dataset; that is, where sources, variables and claims
                 are nodes, and edges connect sources to their claims and
                 claims to their associated variables
        """
        # In this implementation claims must be connected to at least one
        # source, and variables to at least one claim (and in turn to a
        # source), so each connected components contains at least one source.
        # Therefore we may consider which sources are connected to each other,
        # and count the number of partitions.
        #
        # Also note that connected sources are always connected via variables,
        # so we can iterate over sources and keep track of which variables have
        # been seen to determine whether sources are connected
        seen_vars = set([])
        comps = 0
        for row in self.sc:
            # Work out set of IDs for variables this source claims on
            var_ids = {self.claim_ids.inverse[j][0] for j in row.nonzero()[1]}
            # If var_ids does not intersect with the variables seen so far,
            # there is no path in the graph from this sources to a source we
            # have already considered, so we are in a new component
            if not seen_vars.intersection(var_ids):
                comps += 1
            seen_vars.update(var_ids)
        return comps



from truthdiscovery import *

with open("large_synthetic_data.csv") as f:
    sup = SupervisedData.from_csv(f)

alg = TruthFinder(iterator=ConvergenceIterator(DistanceMeasures.L2, 0.001))
res = alg.run(sup.data)
print(res.time_taken)
print(res.iterations)
for s in sorted(res.trust):
    print("{}: {}".format(s, res.trust[s]))
for var in sorted(res.belief):
    bels = res.belief[var]
    for val in sorted(bels):
        print("{}={}: {}".format(var, val, bels[val]))



"""
Script to run various experiments measuring the accuracy of algorithms on
synthetic datasets, and graphing results
"""
from collections import OrderedDict
import itertools
import json
import sys

import numpy as np
import matplotlib.pyplot as plt

from truthdiscovery.input import SyntheticData
from truthdiscovery.algorithm import (
    AverageLog,
    Investment,
    MajorityVoting,
    PooledInvestment,
    Sums,
    TruthFinder
)


ALGORITHMS = OrderedDict({
    "Voting": MajorityVoting(),
    "Sums": Sums(),
    "Average.Log": AverageLog(),
    "Investment": Investment(),
    "Pooled Investment": PooledInvestment(),
    "TruthFinder": TruthFinder()
})


class Experiment:
    # labels for values of independent variable
    labels = None
    # dict mapping algorithm labels to objects
    algorithms = None
    # number of trials to perform for each value
    reps = 10
    # parameters to pass to synthetic data generation. Value for independent
    # variable should be an iterable of values of callables
    synth_params = None

    def __init__(self):
        for key, val in self.synth_params.items():
            # If not an iterable, repeat the value endlessly
            try:
                self.synth_params[key] = iter(val)
            except TypeError:
                self.synth_params[key] = itertools.repeat(val)

    def run(self):
        out = {}
        for var_label in self.labels:
            # results dict: {alg_label: [acc, ...], ...}
            results = {}
            params = dict(self.get_synth_params())
            for _ in range(self.reps):
                this_params = {}
                for key, val in params.items():
                    try:
                        this_params[key] = val()
                    except TypeError:
                        this_params[key] = val

                synth = SyntheticData(**this_params)
                for alg_label, alg in self.algorithms.items():
                    acc = synth.get_accuracy(alg.run(synth.data))
                    this_res = results.setdefault(alg_label, [])
                    this_res.append(acc)
            out[var_label] = results
        return out

    def get_synth_params(self):
        def inner():
            for key, seq in self.synth_params.items():
                val = next(seq)
                yield key, val
        return dict(inner())

    @classmethod
    def run_multiple(cls, *args):
        out = {}
        for exp in args:
            name = exp.__class__.__name__
            print("running {}...".format(name), file=sys.stderr)
            out[name] = exp.run()
        return out

    @classmethod
    def graph(cls, res):
        raise NotImplementedError


class TrustDistExperiment(Experiment):
    algorithms = ALGORITHMS
    labels = None  # set in constructor
    num_sources = 100
    synth_params = {
        "num_variables": 100,
        "claim_probability": 0.1,
        "domain_size": 5,
        "trust": None  # set in constructor
    }

    def __init__(self):
        trust_dists = OrderedDict({
            "Mostly Bad": np.array(list(
                0.75 if i < self.num_sources / 3 else 0.25
                for i in range(self.num_sources)
            )),
            "Uniform": lambda: np.random.uniform(size=(self.num_sources,)),
            r"Normal ($\mu=0.5$, $\sigma=0.15$)": lambda: np.clip(
                np.random.normal(0.5, 0.15, size=(self.num_sources,)),
                0, 1
            ),
        })
        self.labels = trust_dists.keys()
        self.synth_params["trust"] = trust_dists.values()
        super().__init__()

    @classmethod
    def graph(cls, res):
        # Plot trust experiment as bar chart
        fig, ax = plt.subplots()
        num_trust_dists = len(res)
        bar_width = 0.8 / num_trust_dists
        alg_labels = list(cls.algorithms.keys())
        index = np.arange(len(alg_labels))
        for i, (dist, results) in enumerate(res.items()):
            # Get mean accuracy scores for all algorithms for this trust
            # distribution
            acc_scores = list(np.mean(results[alg]) for alg in alg_labels)
            ax.bar(
                index + i * bar_width,
                acc_scores,
                bar_width,
                align="edge",
                label=dist
            )

        ax.set_title(
            "Source trust distribution experiment\n({} sources, {} variables, "
            "{} claim probability, {} domain values)"
            .format(cls.num_sources, cls.synth_params["num_variables"],
                    cls.synth_params["claim_probability"],
                    cls.synth_params["domain_size"])
        )
        ax.set_xlabel("Algorithm")
        ax.set_ylabel("Mean accuracy over {} trials".format(cls.reps))
        ax.set_xticks(index + bar_width * num_trust_dists / 2)
        ax.set_xticklabels(alg_labels)
        ax.legend()
        fig.tight_layout()


class ClaimDensityExperiment(Experiment):
    labels = None  # set in constructor
    algorithms = ALGORITHMS
    synth_params = {
        "trust": lambda: np.random.uniform(size=(100,)),
        "num_variables": 100,
        "claim_probability": None,  # set in constructor
        "domain_size": 10
    }

    def __init__(self):
        claim_probs = list(np.clip(np.arange(0.1, 1.025, 0.05), 0, 1))
        self.synth_params["claim_probability"] = claim_probs
        self.labels = claim_probs
        super().__init__()

    @classmethod
    def graph(cls, res):
        _fig, ax = plt.subplots()
        # JSON does not support non-string keys, so convert to float
        xs = list(map(float, res.keys()))
        for alg in cls.algorithms:
            ys = [np.mean(res[str(x)][alg]) for x in xs]
            ax.plot(xs, ys, "x-", label=alg, linewidth=3)

        ax.set_title(
            "Claim probability experiment\n({} sources, {} variables, {} "
            "domain values)"
            .format(100, cls.synth_params["num_variables"],
                    cls.synth_params["domain_size"])
        )
        ax.set_xlabel("Claim probability")
        ax.set_ylabel("Mean accuracy over {} trials".format(cls.reps))
        ax.legend()


class DomainSizeExperiment(Experiment):
    labels = None
    algorithms = ALGORITHMS
    synth_params = {
        "trust": lambda: np.random.uniform(size=(100,)),
        "num_variables": 100,
        "claim_probability": 0.1,
        "domain_size": None
    }

    def __init__(self):
        domain_sizes = list(range(2, 21))
        self.labels = domain_sizes
        self.synth_params["domain_size"] = domain_sizes
        super().__init__()

    @classmethod
    def graph(cls, res):
        _fig, ax = plt.subplots()
        xs = list(map(int, res.keys()))
        for alg in cls.algorithms:
            ys = [np.mean(res[str(x)][alg]) for x in xs]
            ax.plot(xs, ys, "x-", label=alg, linewidth=3)

        ax.set_title(
            "Domain size experiment\n({} sources, {} variables, {} "
            "claim probability)"
            .format(100, cls.synth_params["num_variables"],
                    cls.synth_params["claim_probability"])
        )
        ax.set_xlabel("Variable domain size")
        ax.set_ylabel("Mean accuracy over {} trials".format(cls.reps))
        ax.legend()


def main():
    if "-h" in sys.argv or "--help" in sys.argv:
        print("usage: {}".format(sys.argv[0]))
        print("       {} graph FILE\n".format(sys.argv[0]))
        print("in the first form, run the experiment and print results as "
              "JSON to stdout")
        print("in the second, graph the results from the given file")
        return

    if len(sys.argv) >= 3 and sys.argv[1] == "graph":
        with open(sys.argv[2]) as json_file:
            res = json.load(json_file)
        for cls_name, results in res.items():
            globals()[cls_name].graph(results)
        plt.show()
        return

    res = Experiment.run_multiple(
        # ClaimDensityExperiment(),
        # TrustDistExperiment(),
        DomainSizeExperiment()
    )
    print(json.dumps(res, indent=2))


if __name__ == "__main__":
    main()



class ConvergenceError(Exception):
    """
    An algorithm failed to converge within the iteration limit
    """


class EarlyFinishError(Exception):
    """
    Error raised when an iterative algorithm is forced to finish early, e.g.
    due to rounding errors preventing further computations
    """


class EmptyDatasetError(Exception):
    """
    An algorithm was run on a dataset containing no claims
    """



class Entity:
    """
    Class to represent an abstract entity that can be drawn
    """
    def __init__(self, x=None, y=None, colour=None):
        self.x = x
        self.y = y
        self.colour = colour


class Rectangle(Entity):
    def __init__(self, width=None, height=None, **kwargs):
        super().__init__(**kwargs)
        self.width = width
        self.height = height


class Circle(Entity):
    def __init__(self, radius=None, label=None, **kwargs):
        super().__init__(**kwargs)
        self.radius = radius
        self.label = label


class Line(Entity):
    def __init__(self, end_x=None, end_y=None, width=None, dashed=False,
                 **kwargs):
        super().__init__(**kwargs)
        self.end_x = end_x
        self.end_y = end_y
        self.width = width
        self.dashed = dashed


class Label(Entity):
    def __init__(self, text=None, size=None, overflow_background=None,
                 **kwargs):
        """
        :param text:   text to draw
        :param size:   size (backend specific units)
        :param overflow_background: if the label is associated with another
                                    entity and exceeds its bounds, draw the
                                    label in a box with this background
        """
        # Note: coordinates of label are the coordinates of the *centre*
        super().__init__(**kwargs)
        self.text = text
        self.size = size
        self.overflow_background = overflow_background



"""
Command-line interface to truthdiscovery library
"""
import argparse
import sys

import numpy as np
import yaml

from truthdiscovery.client.base import BaseClient, OutputFields
from truthdiscovery.client.web import run_debug_server
from truthdiscovery.input import MatrixDataset, SupervisedData, SyntheticData
from truthdiscovery.graphs import MatrixDatasetGraphRenderer


def numpy_float_yaml_representer(dumper, val):
    """
    Convert numpy float64 objects to Python floats for YAML representation
    """
    return dumper.represent_float(float(val))


yaml.add_representer(np.float64, numpy_float_yaml_representer)


class CommandLineClient(BaseClient):
    def make_argparse_type(self, method):
        """
        Decorate a function to catch any ValueError and convert it to
        argparse's ArgumentTypeError
        """
        def inner(*args, **kwargs):
            try:
                return method(*args, **kwargs)
            except ValueError as ex:
                raise argparse.ArgumentTypeError(ex)
        # Function name is used in argparse's error message, so update it
        inner.__name__ = method.__name__
        return inner

    def get_parser(self):
        """
        :return: argparse ArgumentParser object
        """
        parser = argparse.ArgumentParser(description=__doc__)
        subparsers = parser.add_subparsers(
            dest="command",
            metavar="COMMAND",
        )
        # Run sub-command
        run_parser = subparsers.add_parser(
            "run",
            help="Run truth discovery algorithms on a CSV dataset",
            description="""
                Run algorithms on a dataset loaded from a CSV file, and
                return results in YAML format
            """,
        )
        run_parser.add_argument(
            "-a", "--algorithm",
            help="Algorithms to run: choose from {}".format(
                ", ".join(self.ALG_LABEL_MAPPING.keys())
            ),
            required=True,
            dest="alg_classes",
            metavar="ALGORITHM",
            type=self.make_argparse_type(self.algorithm_cls),
            nargs="+",
        )
        run_parser.add_argument(
            "-p", "--params",
            help=("""
                Parameters to pass algorithms, each in the form 'key=value'.
                Unkown parameters are ignored. For 'priors', see the
                PriorBelief enumeration for valid values. For 'iterator', use
                the format 'fixed-<N>' for fixed N iterations, or
                '<measure>-convergence-<threshold>[-limit-<N>]' for convergence
                in 'measure' within 'threshold', up to an optional maximum
                number 'limit' iterations.
            """),
            dest="alg_params",
            metavar="PARAM",
            type=self.make_argparse_type(self.algorithm_parameter),
            nargs="+",
        )
        run_parser.add_argument(
            "-f", "--dataset",
            help="CSV file to run the algorithm on",
            type=argparse.FileType("r"),
            required=True
        )
        run_parser.add_argument(
            "-s", "--supervised",
            help=("Interpret the CSV as a supervised dataset, i.e. treat the "
                  "first row as known true values"),
            action="store_true",
            dest="supervised"
        )
        # Output options
        output_group = run_parser.add_argument_group(
            title="output options",
            description="options for controlling what output is given"
        )
        output_group.add_argument(
            "--sources",
            help=("Sources to restrict results to. Unknown sources are "
                  "ignored"),
            dest="sources",
            metavar="SOURCE",
            nargs="+",
            type=int
        )
        output_group.add_argument(
            "--variables",
            help=("Variables to restrict results to. Unknown variables are "
                  "ignored"),
            dest="variables",
            metavar="VAR",
            nargs="+",
            type=int
        )
        output_field_choices = [f.value for f in OutputFields]
        output_group.add_argument(
            "-o", "--output",
            help=(
                "Fields to include in the output: choose from {}"
                .format(", ".join(output_field_choices))
            ),
            metavar="OUTPUT_FIELD",
            dest="output_fields",
            nargs="+",
            type=OutputFields,
            default=[
                OutputFields.TRUST, OutputFields.BELIEF,
                OutputFields.ITERATIONS, OutputFields.TIME
            ]
        )

        # Synthetic data generation sub-command
        synth_parser = subparsers.add_parser(
            "synth",
            help="Generate a synthetic CSV dataset",
            description="""
                Randomly generate a CSV dataset based on a given list of source
                trust scores, which are interpreted as the probability that
                each source makes a correct claim. The first row in the output
                gives the true values, and the subsequent rows are the source
                claims.
            """
        )
        synth_parser.add_argument(
            "--trust",
            help="Trust scores for sources in [0, 1]",
            metavar="TRUST_SCORE",
            nargs="+",
            required=True,
            type=float
        )
        synth_parser.add_argument(
            "--num-vars",
            help="The number of variables to generate",
            metavar="NUM_VARS",
            type=int
        )
        synth_parser.add_argument(
            "--claim-prob",
            help=("The probability that a given source will make a claim for "
                  "a given variable"),
            metavar="CLAIM_PROB",
            type=float
        )
        synth_parser.add_argument(
            "--domain-size",
            help="The number of possible values for each variable",
            metavar="DOMAIN_SIZE",
            type=int
        )
        # Graph generation sub-command
        graph_parser = subparsers.add_parser(
            "graph",
            help="Generate a graph representation of a dataset as a PNG",
            description="Generate a graph representation of a dataset as a PNG"
        )
        graph_parser.add_argument(
            "-f", "--dataset",
            help="CSV file to create a graph for",
            type=argparse.FileType("r"),
            required=True
        )
        graph_parser.add_argument(
            "-o", "--outfile",
            help="Path to save output PNG to",
            type=argparse.FileType("wb"),
            required=True
        )
        # Optional renderer parameters
        graph_parser.add_argument(
            "--width",
            help="Width in pixels",
            type=int
        )
        graph_parser.add_argument(
            "--node-radius",
            help="Radius of each node in pixels",
            type=int
        )
        graph_parser.add_argument(
            "--spacing",
            help="(minimum) vertical spacing between nodes in pixels",
            type=int
        )
        graph_parser.add_argument(
            "--font-size",
            dest="font_size",
            help="Font size for node labels",
            type=int
        )
        graph_parser.add_argument(
            "--line-width",
            dest="line_width",
            help="Width of edges in pixels",
            type=int
        )
        graph_parser.add_argument(
            "--node-border-width",
            dest="node_border_width",
            help="Width node borders in pixels",
            type=int
        )
        graph_parser.add_argument(
            "--one-indexed",
            dest="one_indexed",
            help="Start source/variable numbering from 1 instead of 0",
            action="store_true"
        )
        # HTTP server sub-parser
        _httpd_parser = subparsers.add_parser(
            "httpd",
            help=("Start the debug Flask server for a web-interface to the "
                  "library"),
            description="""
                Start the debug Flask server for the HTTP API and web-interface
                to the truthdiscovery library
            """,
        )
        return parser

    def run(self, cli_args):
        parser = self.get_parser()
        args = parser.parse_args(cli_args)

        if args.command == "run":
            self.run_algorithm(args, parser)
        elif args.command == "synth":
            self.generate_synthetic(args, parser)
        elif args.command == "graph":
            self.generate_graph(args, parser)
        elif args.command == "httpd":  # pragma: no cover
            run_debug_server()
        else:
            parser.print_help()

    def run_algorithm(self, args, parser):
        alg_objs = []
        all_params = dict(args.alg_params or [])
        for cls in args.alg_classes:
            params, ignored = self.get_algorithm_params(cls, all_params)
            alg_objs.append(self.get_algorithm_object(cls, params))

            if ignored:
                msg = self.get_ignored_parameters_message(cls, ignored)
                print("WARNING: {}".format(msg), file=sys.stderr)

        sup_data = None
        dataset = None
        if args.supervised:
            sup_data = SupervisedData.from_csv(args.dataset)
            dataset = sup_data.data
        else:
            # Catch error early if accuracy requested in output but dataset is
            # not supervised
            if OutputFields.ACCURACY in args.output_fields:
                parser.error("cannot calculate accuracy without --supervised")
            dataset = MatrixDataset.from_csv(args.dataset)

        output_obj = {}
        for alg in alg_objs:
            results = alg.run(dataset).filter(sources=args.sources,
                                              variables=args.variables)

            # Get results to display
            label = self.ALG_LABEL_MAPPING.inverse[alg.__class__]
            output_obj[label] = self.get_output_obj(
                results,
                output_fields=args.output_fields,
                sup_data=sup_data
            )
        print(yaml.dump(output_obj, indent=2, default_flow_style=False))

    def generate_synthetic(self, args, parser):
        kwargs = {
            "trust": args.trust,
            "num_variables": args.num_vars,
            "claim_probability": args.claim_prob,
            "domain_size": args.domain_size
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        try:
            print(SyntheticData(**kwargs).to_csv())
        except ValueError as ex:
            parser.error(ex)

    def get_graph_renderer(self, args):
        """
        :param args:    argparse params
        :return:        a any:`GraphRenderer` object
        """
        kwargs = {
            "width": args.width,
            "node_radius": args.node_radius,
            "spacing": args.spacing,
            "font_size": args.font_size,
            "line_width": args.line_width,
            "node_border_width": args.node_border_width,
            "zero_indexed": not args.one_indexed
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return MatrixDatasetGraphRenderer(**kwargs)

    def generate_graph(self, args, parser):
        try:
            renderer = self.get_graph_renderer(args)
            dataset = MatrixDataset.from_csv(args.dataset)
        except ValueError as ex:  # pragma: no cover
            parser.error(ex)
        renderer.render(dataset, args.outfile)


def main():  # pragma: no cover
    client = CommandLineClient()
    client.run(sys.argv[1:])



import json

from truthdiscovery.graphs.backends.base import BaseBackend
from truthdiscovery.graphs.entities import Entity


class JsonBackend(BaseBackend):
    """
    Serialise entities to draw as a JSON document
    """
    def draw_entities(self, entities, outfile, width, height):
        obj = {
            "width": width,
            "height": height,
            "entities": [JsonBackend.entity_to_dict(e) for e in entities]
        }
        json.dump(obj, outfile)

    @classmethod
    def entity_to_dict(cls, entity):
        """
        :return: a dict representation of an :any:`Entity` object
        """
        if not isinstance(entity, Entity):
            return entity

        obj = {"type": entity.__class__.__name__.lower()}
        for name, val in entity.__dict__.items():
            obj[name] = cls.entity_to_dict(val)
        return obj



import requests
import textacy
import spacy


NEWSAPI_KEY = "0797893c0ac04d8bbd4e89d43cb27539"
TEXTS = None


def get_claims(term=None, limit=100):
    print("getting articles...")
    base_url = "https://newsapi.org/v2/top-headlines"
    params = {
        "sortBy": "popularity",
        "language": "en",
        "country": "gb",
        "category": "business",
        "apiKey": NEWSAPI_KEY
    }
    if term is not None:
        params["q"] = term
    print(params)
    resp = requests.get(base_url, params=params)
    results = resp.json()
    try:
        articles = results["articles"]
    except KeyError:
        print(resp.json())
        return []
    for article in articles[:limit]:
        yield article["title"]
        yield article["description"]
        yield article["content"]


def main_textacy():
    # Make sure to download an English language model with
    # `python -m spacy download en_core_web_sm`
    en = textacy.load_spacy("en_core_web_sm")

    for text in TEXTS:
        doc = textacy.Doc(text, lang=en)
        triples = textacy.extract.subject_verb_object_triples(doc)
        if triples:
            print("original text:")
            print(text)
            print("subject, verb, object:")
            for subject, verb, obj in triples:
                print("s: '{}', v: '{}', o: '{}'".format(subject, verb, obj))
            print("")


def main_spacy():
    nlp = spacy.load("en_core_web_sm")
    for text in TEXTS:
        print("original text:")
        print("  ", text)
        doc = nlp(text)

        for sent in doc.sents:
            subjects = [w for w in sent if w.dep_ == "nsubj"]
            for subj in subjects:
                print("subject: {}".format(subj))
                print("lefts: {}".format(list(subj.lefts)))


if __name__ == "__main__":
    # with open("texts.txt") as f:
    #     TEXTS = f.read().split("\n")
    TEXTS = get_claims()

    # main = main_textacy
    main = main_spacy
    main()



from truthdiscovery.utils.iterator import (
    ConvergenceIterator,
    DistanceMeasures,
    FixedIterator,
    Iterator
)


def filter_dict(dct, keys):
    """
    :yield: Tuples ``(key, value)`` for keys in ``keys``
    """
    for key in keys:
        try:
            yield (key, dct[key])
        except KeyError:
            pass



from enum import Enum

import numpy as np

from truthdiscovery.exceptions import ConvergenceError


class DistanceMeasures(Enum):
    """
    Enumeration of possible methods for calculating the distance between
    vectors. Note that not all methods are actually distance measures in the
    sense of metric spaces.
    """
    #: L1 norm (Manhattan distance)
    L1 = "l1"
    #: L2 norm (Euclidean distance)
    L2 = "l2"
    #: Infinity norm
    L_INF = "l_inf"
    #: 1 - cosine similarity
    COSINE = "cosine"


class Iterator:
    """
    Base class for iterators
    """
    it_count = 0

    def compare(self, _obj1, _obj2):
        """
        Log the comparison of two objects
        """
        # By default just increase iteration count without comparing objects at
        # all
        self.it_count += 1

    def finished(self):
        """
        Return True if iteration is finished, False otherwise
        """
        raise NotImplementedError("Must be implemented in child classes")

    def reset(self):
        """
        Reset iteration
        """
        self.it_count = 0


class FixedIterator(Iterator):
    """
    Iterator that runs a fixed number of times, then finishes
    """
    limit = 20

    def __init__(self, limit=None):
        """
        :param limit: number of iterations to perform
        """
        if limit is not None:
            if limit < 0:
                raise ValueError("Iteration limit cannot be negative")
            self.limit = limit
        self.reset()

    def finished(self):
        return self.it_count >= self.limit


class ConvergenceIterator(Iterator):
    """
    Iterator that runs until the distance between vectors of interest is
    smaller than a set threshold
    """
    threshold = 0.0001
    distance_measure = None
    current_distance = None
    limit = 1000000
    debug = False

    def __init__(self, distance_measure, threshold, limit=None, debug=False):
        """
        :param distance_measure: value from :any:`DistanceMeasures` enumeration
        :param threshold:        iteration is finished when distance goes below
                                 this threshold value
        :param limit:            upper limit on number of iterations to perform
        :param debug:            if True, print out current distance at each
                                 iteration
        """
        self.distance_measure = distance_measure
        if threshold is not None:
            self.threshold = threshold
        if limit is not None:
            self.limit = limit
        self.debug = debug
        self.reset()

    def reset(self):
        super().reset()
        self.current_distance = None

    def compare(self, obj1, obj2):
        """
        Update the most recent distance between objects
        """
        super().compare(obj1, obj2)
        self.current_distance = self.get_distance(
            self.distance_measure, obj1, obj2
        )
        if self.debug:  # pragma: no cover
            print("{},{}".format(self.it_count, self.current_distance))

    def finished(self):
        """
        :raises ConvergenceError: if maximum iteration limit has been reached
                                  without distance going below threshold
        """
        if (self.current_distance is not None
                and self.current_distance < self.threshold):
            return True
        if self.it_count >= self.limit:
            raise ConvergenceError(
                "Did not converge in {} iterations".format(self.limit)
            )
        return False

    @classmethod
    def get_distance(cls, distance_measure, obj1, obj2):
        """
        Calculate distance between vectors using the given measure

        :param distance_measure: value from :any:`DistanceMeasures` enumeration
        :param obj1:             first object to be compared
        :param obj2:             second object to be compared
        :raises ValueError: if ``distance_measure`` is not an item from the
                            :any:`DistanceMeasures` enumeration
        """
        if distance_measure == DistanceMeasures.L1:
            return np.linalg.norm(obj1 - obj2, ord=1)
        if distance_measure == DistanceMeasures.L2:
            return np.linalg.norm(obj1 - obj2)
        if distance_measure == DistanceMeasures.L_INF:
            return np.linalg.norm(obj1 - obj2, ord=np.inf)
        if distance_measure == DistanceMeasures.COSINE:
            norm1 = np.linalg.norm(obj1)
            norm2 = np.linalg.norm(obj2)
            if norm1 == 0 or norm2 == 0:
                return 1
            return np.clip(1 - (np.dot(obj1, obj2) / (norm1 * norm2)), 0, 1)
        raise ValueError(
            "Invalid distance measure: '{}'".format(distance_measure)
        )



import sys

import numpy.ma as ma

from truthdiscovery.algorithm import Investment
from truthdiscovery.input import MatrixDataset
from truthdiscovery.graphs import (
    GifAnimator,
    MatrixDatasetGraphRenderer,
    ResultsGradientColourScheme
)
from truthdiscovery.utils import ConvergenceIterator, DistanceMeasures


if __name__ == "__main__":
    try:
        out_path = sys.argv[1]
    except IndexError:
        print("usage: {} DEST".format(sys.argv[0]), file=sys.stderr)
        sys.exit(1)

#     tuples = [
#         ("source 1", "x", 4),
#         ("source 1", "y", 7),
#         ("source 2", "y", 7),
#         ("source 2", "z", 5),
#         ("source 3", "x", 3),
#         ("source 3", "z", 5),
#         ("source 4", "x", 3),
#         ("source 4", "y", 6),
#         ("source 4", "z", 8)
#     ]
#     mydata = Dataset(tuples)

    mydata = MatrixDataset(ma.masked_values([
        [1, 9, 3, 4],
        [2, 2, 9, 9],
        [9, 9, 7, 9],
        [1, 2, 5, 9]
    ], 9))

    it = ConvergenceIterator(DistanceMeasures.L2, 0.001)
    algorithm = Investment(iterator=it)

    cs = ResultsGradientColourScheme(algorithm.run(mydata))
    rend = MatrixDatasetGraphRenderer(zero_indexed=False, colours=cs)
    animator = GifAnimator(renderer=rend, frame_duration=0.2)
    with open(out_path, "wb") as outfile:
        animator.animate(outfile, algorithm, mydata)



import json

from flask import Flask
import numpy.ma as ma
import pytest
import yaml

from truthdiscovery.algorithm import (
    AverageLog,
    MajorityVoting,
    PooledInvestment,
    PriorBelief,
    Sums,
    TruthFinder
)
from truthdiscovery.client import BaseClient, CommandLineClient, OutputFields
from truthdiscovery.client.web import get_flask_app, route
from truthdiscovery.input import MatrixDataset, SupervisedData
from truthdiscovery.utils import (
    ConvergenceIterator,
    DistanceMeasures,
    FixedIterator
)
from truthdiscovery.test.utils import is_valid_png


class ClientTestsBase:
    @pytest.fixture
    def dataset(self):
        return MatrixDataset(ma.masked_values([
            [1, 2, 3, 2],
            [3, 0, 1, 2],
            [2, 2, 0, 0],
            [0, 1, 0, 3]
        ], 0))

    @pytest.fixture
    def csv_fileobj(self, dataset, tmpdir):
        csvfile = tmpdir.join("data.csv")
        csvfile.write(dataset.to_csv())
        return csvfile

    @pytest.fixture
    def csv_dataset(self, csv_fileobj):
        return str(csv_fileobj)


class TestBaseClient(ClientTestsBase):
    def test_get_iterator(self):
        fixed_145 = BaseClient().get_iterator("fixed-145")
        assert isinstance(fixed_145, FixedIterator)
        assert fixed_145.limit == 145

        l1_conv = BaseClient().get_iterator("l1-convergence-0.234")
        assert isinstance(l1_conv, ConvergenceIterator)
        assert l1_conv.distance_measure == DistanceMeasures.L1
        assert l1_conv.threshold == 0.234

        l2_with_limit = BaseClient().get_iterator(
            "l2-convergence-0.234-limit-9"
        )
        assert isinstance(l2_with_limit, ConvergenceIterator)
        assert l2_with_limit.distance_measure == DistanceMeasures.L2
        assert l2_with_limit.threshold == 0.234
        assert l2_with_limit.limit == 9

        # Should be too many iterations
        with pytest.raises(ValueError):
            fixed_2000 = BaseClient().get_iterator(
                "fixed-2000", max_limit=1999
            )
        with pytest.raises(ValueError):
            conv_2000 = BaseClient().get_iterator(
                "l2-convergence-1-limit-2000", max_limit=1999
            )

        invalid_it_strings = (
            "fixed",
            "fixed-",
            "fixed-hello",
            "fixed-10.0",
            "fixed--4",
            "convergence",
            "-convergence-",
            "l1-convergence-",
            "l1-convergence-l1",
            "blah-convergence-0.03",
            "l1-convergence-0.03-limit",
            "l1-convergence-0.03-limit-",
            "l1-convergence-0.03-limit-45.0"
        )
        for it_string in invalid_it_strings:
            with pytest.raises(ValueError):
                BaseClient().get_iterator(it_string)

    def test_get_algorithm_parameter(self):
        # Iterator param
        name1, val1 = BaseClient().algorithm_parameter("iterator=fixed-99")
        assert name1 == "iterator"
        assert isinstance(val1, FixedIterator)
        assert val1.limit == 99

        # Priors
        name2, val2 = BaseClient().algorithm_parameter("priors=voted")
        assert name2 == "priors"
        assert val2 == PriorBelief.VOTED

        # Anything else should be a float
        name3, val3 = BaseClient().algorithm_parameter("g=1.4")
        assert name3 == "g"
        assert val3 == 1.4

        # Extra whitespace should be allowed
        name3, val3 = BaseClient().algorithm_parameter("ppp =   3.4")
        assert name3 == "ppp"
        assert val3 == 3.4

    def test_get_output_obj(self, csv_fileobj):
        dataset = MatrixDataset.from_csv(csv_fileobj)
        alg = Sums(iterator=FixedIterator(5))
        # Default should be all fields if none are given, but not accuracy
        # unless supervised data given
        results = alg.run(dataset)
        out1 = BaseClient().get_output_obj(results)
        exp_keys = {
            f.value for f in OutputFields if f != OutputFields.ACCURACY
        }
        assert set(out1.keys()) == exp_keys

        sup_data = SupervisedData.from_csv(csv_fileobj)
        sup_results = alg.run(sup_data.data)
        out2 = BaseClient().get_output_obj(sup_results, sup_data=sup_data)
        assert set(out2.keys()) == {f.value for f in OutputFields}
        assert out2["trust"] == sup_results.trust
        assert out2["belief"] == sup_results.belief

        out3 = BaseClient().get_output_obj(
            results, output_fields=[OutputFields.TRUST]
        )
        assert set(out3.keys()) == {"trust"}

    def test_get_algorithm_params(self):
        client = BaseClient()
        params = {
            "priors": PriorBelief.VOTED, "g": 1.99,
            "dampening_factor": 0.5678
        }

        def get_p(*args):
            return client.get_algorithm_params(*args)[0]

        assert get_p(Sums, params) == {"priors": PriorBelief.VOTED}
        assert get_p(TruthFinder, params) == {
            "priors": PriorBelief.VOTED, "dampening_factor": 0.5678
        }
        assert get_p(PooledInvestment, params) == {
            "priors": PriorBelief.VOTED, "g": 1.99
        }
        assert get_p(MajorityVoting, params) == {}


class TestCommandLineClient(ClientTestsBase):
    @pytest.fixture
    def client(self):
        return CommandLineClient()

    def get_algorithm_objects(self, client, args):
        all_params = dict(args.alg_params or [])
        objs = []
        for cls in args.alg_classes:
            params, _ignored = client.get_algorithm_params(cls, all_params)
            obj = client.get_algorithm_object(cls, params)
            objs.append(obj)
        return objs

    def run(self, *args):
        client = CommandLineClient()
        client.run(args)

    def get_parsed_args(self, *args):
        return CommandLineClient().get_parser().parse_args(args)

    def test_no_commands(self, capsys):
        self.run()
        out = capsys.readouterr().out
        assert "Command-line interface to truthdiscovery library" in out

    def test_basic(self, csv_dataset):
        self.run(
            "run", "--algorithm", "sums", "-f", csv_dataset
        )

    def test_results(self, csv_dataset, csv_fileobj, capsys):
        self.run(
            "run", "-a", "average_log", "-f", csv_dataset
        )
        got_results = yaml.safe_load(capsys.readouterr().out)
        assert "average_log" in got_results
        alg_results = got_results["average_log"]
        assert isinstance(alg_results, dict)
        exp_results = AverageLog().run(MatrixDataset.from_csv(csv_fileobj))
        assert alg_results["trust"] == exp_results.trust
        assert alg_results["belief"] == exp_results.belief
        assert alg_results["iterations"] == exp_results.iterations

    def test_multiple_parameters(self, csv_dataset):
        self.run(
            "run", "--algorithm", "pooled_investment", "-p", "g=1.1234",
            "iterator=fixed-11", "priors=uniform", "-f", csv_dataset
        )

    def test_multiple_algorithms(self, capsys, csv_dataset):
        # Mirrors an identical test for web client below
        self.run(
            "run", "-a", "investment", "sums", "voting", "-f", csv_dataset,
            "-p", "iterator=fixed-11", "g=1.16"
        )
        outerr = capsys.readouterr()
        got_results = yaml.safe_load(outerr.out)

        assert set(got_results.keys()) == {"investment", "sums", "voting"}
        assert got_results["investment"]["iterations"] == 11
        assert got_results["sums"]["iterations"] == 11
        assert got_results["voting"]["iterations"] is None

        assert got_results["voting"]["trust"] == {i: 1 for i in range(4)}
        assert got_results["sums"]["trust"] == {
            0: 1.0, 1: 0.5323216608830517, 2: 0.3472167514190226,
            3: 5.9742483641484065e-05
        }
        assert got_results["investment"]["trust"] == {
            0: 1.0, 1: 0.8725062856135304, 2: 0.6845094785538429,
            3: 0.1211362221507375
        }

        # Should get warnings for the invalid parameters
        expected_warnings = {
            "WARNING: Ignored parameters for 'MajorityVoting': g, iterator",
            "WARNING: Ignored parameters for 'Sums': g"
        }
        assert set(filter(None, outerr.err.split("\n"))) == expected_warnings

    def test_get_algorithm_instance(self, client, csv_dataset):
        args = self.get_parsed_args(
            "run", "--algorithm", "voting", "truthfinder", "-p",
            "dampening_factor=0.1", "influence_param=0.77", "-f", csv_dataset
        )
        voting = self.get_algorithm_objects(client, args)[0]
        assert isinstance(voting, MajorityVoting)
        tf = self.get_algorithm_objects(client, args)[1]
        assert isinstance(tf, TruthFinder)
        assert tf.dampening_factor == 0.1
        assert tf.influence_param == 0.77

    def test_set_prior_belief(self, client, csv_dataset, capsys):
        args = self.get_parsed_args(
            "run", "--algorithm", "sums", "-p", "priors=voted", "-f",
            csv_dataset
        )
        alg = self.get_algorithm_objects(client, args)[0]
        assert isinstance(alg, Sums)
        assert alg.priors == PriorBelief.VOTED

        # Invalid prior string
        with pytest.raises(SystemExit):
            self.run(
                "run", "--algorithm", "sums", "-p", "priors=blah", "-f",
                csv_dataset
            )

        err_msg = capsys.readouterr().err
        assert "'blah' is not a valid PriorBelief" in err_msg

    def test_set_iterator(self, client, csv_dataset, capsys):
        # Fixed iterator
        raw_args1 = (
            "run", "--algorithm", "sums", "-p", "iterator=fixed-123", "-f",
            csv_dataset
        )
        args1 = self.get_parsed_args(*raw_args1)
        alg1 = self.get_algorithm_objects(client, args1)[0]
        assert isinstance(alg1, Sums)
        assert isinstance(alg1.iterator, FixedIterator)
        assert alg1.iterator.limit == 123
        self.run(*raw_args1)
        results = yaml.safe_load(capsys.readouterr().out)["sums"]
        assert results["iterations"] == 123

        # Convergence iterator
        args2 = self.get_parsed_args(
            "run", "--algorithm", "sums", "-p",
            "iterator=cosine-convergence-0.01", "-f", csv_dataset
        )
        alg2 = self.get_algorithm_objects(client, args2)[0]
        assert isinstance(alg2.iterator, ConvergenceIterator)
        assert alg2.iterator.distance_measure == DistanceMeasures.COSINE
        assert alg2.iterator.threshold == 0.01

        # Convergence iterator with limit
        args3 = self.get_parsed_args(
            "run", "--algorithm", "sums", "-p",
            "iterator=l2-convergence-0.3-limit-99", "-f", csv_dataset
        )
        alg3 = self.get_algorithm_objects(client, args3)[0]
        assert isinstance(alg3.iterator, ConvergenceIterator)
        assert alg3.iterator.distance_measure == DistanceMeasures.L2
        assert alg3.iterator.threshold == 0.3
        assert alg3.iterator.limit == 99

        # Convergence iterator with invalid distance measure
        with pytest.raises(SystemExit):
            self.run(
                "run", "--algorithm", "sums", "-p",
                "iterator=blah-convergence-0.3-limit-99", "-f", csv_dataset
            )
        assert "invalid distance measure 'blah'" in capsys.readouterr().err

        # Invalid iterator specification
        with pytest.raises(SystemExit):
            self.run(
                "run", "--algorithm", "sums", "-p", "iterator=hello", "-f",
                csv_dataset
            )
        assert "invalid iterator specification" in capsys.readouterr().err

    def test_invalid_algorithm_parameter(self, csv_dataset, capsys):
        # Invalid format
        with pytest.raises(SystemExit):
            self.run(
                "run", "--algorithm", "truthfinder", "-p", "initial_trust 0.1",
                "-f", csv_dataset
            )
        assert "must be in the form 'key=value'" in capsys.readouterr().err

    def test_invalid_algorithm(self, csv_dataset, capsys):
        with pytest.raises(SystemExit):
            self.run(
                "run", "--algorithm", "joesalgorithm", "-f", csv_dataset
            )
        assert (
            "invalid algorithm label 'joesalgorithm'"
            in capsys.readouterr().err
        )

    def test_filter_sources_variables(self, csv_dataset, capsys):
        # Filter sources
        self.run(
            "run", "-a", "sums", "-f", csv_dataset, "--sources", "0", "3",
        )
        results1 = yaml.safe_load(capsys.readouterr().out)["sums"]
        assert set(results1["trust"].keys()) == {0, 3}
        assert set(results1["belief"].keys()) == {0, 1, 2, 3}

        # Filter both
        self.run(
            "run", "-a", "sums", "-f", csv_dataset, "--sources", "0", "3",
            "--variables", "1", "2"
        )
        results2 = yaml.safe_load(capsys.readouterr().out)["sums"]
        assert set(results2["trust"].keys()) == {0, 3}
        assert set(results2["belief"].keys()) == {1, 2}

        # Special case where only one source/variable
        self.run(
            "run", "-a", "sums", "-f", csv_dataset, "--sources", "1",
            "--variables", "0"
        )
        results3 = yaml.safe_load(capsys.readouterr().out)["sums"]
        assert set(results3["trust"].keys()) == {1}
        assert set(results3["belief"].keys()) == {0}

        # Unknown sources/vars should not cause trouble
        self.run(
            "run", "-a", "sums", "-f", csv_dataset, "--sources", "3", "1000",
            "--variables", "499", "666"
        )
        results3 = yaml.safe_load(capsys.readouterr().out)["sums"]
        assert set(results3["trust"].keys()) == {3}
        # We didn't give any valid variables: belief should be empty
        assert results3["belief"] == {}

    def test_default_output(self, csv_dataset, capsys):
        self.run("run", "-a", "voting", "-f", csv_dataset)
        results = yaml.safe_load(capsys.readouterr().out)["voting"]
        assert set(results.keys()) == {
            "time", "iterations", "trust", "belief"
        }
        assert results["iterations"] is None

    def test_custom_output(self, csv_fileobj, csv_dataset, capsys):
        self.run("run", "-a", "sums", "-f", csv_dataset, "-o", "time")
        results = yaml.safe_load(capsys.readouterr().out)["sums"]
        assert set(results.keys()) == {"time"}

        self.run(
            "run", "-a", "sums", "-f", csv_dataset, "-o", "time",
            "iterations"
        )
        results = yaml.safe_load(capsys.readouterr().out)["sums"]
        assert set(results.keys()) == {"time", "iterations"}

        self.run(
            "run", "-a", "sums", "-f", csv_dataset, "-o", "trust",
            "trust_stats"
        )
        results = yaml.safe_load(capsys.readouterr().out)["sums"]
        assert set(results.keys()) == {"trust", "trust_stats"}
        exp_mean, exp_stddev = (Sums().run(MatrixDataset.from_csv(csv_fileobj))
                                .get_trust_stats())
        assert results["trust_stats"] == {
            "mean": exp_mean, "stddev": exp_stddev
        }

    def test_show_most_believed_values(self, csv_dataset, capsys):
        self.run(
            "run", "-a", "voting", "-f", csv_dataset, "--output",
            "most_believed_values"
        )
        results = yaml.safe_load(capsys.readouterr().out)["voting"]
        assert results == {
            "most_believed_values": {0: [1, 2, 3], 1: [2], 2: [1, 3], 3: [2]}
        }
        # Test with variable filtering
        self.run(
            "run", "-a", "voting", "-f", csv_dataset, "-o",
            "most_believed_values", "--variables", "0", "3"
        )
        results = yaml.safe_load(capsys.readouterr().out)["voting"]
        assert "belief" not in results
        assert "most_believed_values" in results
        assert results["most_believed_values"] == {
            0: [1, 2, 3],
            3: [2]
        }

    def test_belief_stats(self, csv_dataset, csv_fileobj, capsys):
        self.run("run", "-a", "sums", "-f", csv_dataset, "-o", "belief_stats")
        results = yaml.safe_load(capsys.readouterr().out)["sums"]
        assert set(results.keys()) == {"belief_stats"}
        exp_belief_stats = (Sums().run(MatrixDataset.from_csv(csv_fileobj))
                            .get_belief_stats())
        assert results["belief_stats"] == {
            var: {"mean": mean, "stddev": stddev}
            for var, (mean, stddev) in exp_belief_stats.items()
        }

    def test_synthetic_generation(self, capsys):
        self.run(
            "synth", "--trust", "0.5", "0.6", "0.7", "--num-vars", "10",
            "--domain-size", "5"
        )
        output = capsys.readouterr().out.strip()
        lines = output.split("\n")
        assert len(lines) == 4  # first row is true values, then 3 source rows
        for line in lines:
            columns = line.split(",")
            assert len(columns) == 10
            for col in columns:
                assert col == "" or float(col) in {0, 1, 2, 3, 4}

    def test_synthetic_generation_claim_prob_1(self, capsys):
        self.run(
            "synth", "--trust", "0.5", "0.6", "0.7", "--num-vars", "10",
            "--domain-size", "2", "--claim-prob", "1"
        )
        output = capsys.readouterr().out.strip()
        lines = output.split("\n")
        assert len(lines) == 4
        for line in lines:
            columns = line.split(",")
            for col in columns:
                assert float(col) in {0, 1}

    def test_synthetic_generation_source_trust_1(self, capsys):
        self.run("synth", "--trust", "1", "--claim-prob", "1")
        output = capsys.readouterr().out.strip()
        true_vals, claims = output.split("\n")
        assert true_vals == claims

    def test_synthetic_generation_invalid_params(self, capsys):
        # Check that invalid param errors are caught by the parser, not raised
        # as Python exceptions
        with pytest.raises(SystemExit):
            self.run("synth", "--trust", "2")
        exp_err_msg = "error: Trust values must be in [0, 1]"
        assert exp_err_msg in capsys.readouterr().err

    def test_supervised_dataset_and_accuracy(self, csv_dataset, capsys):
        self.run(
            "run", "-a", "voting", "-f", csv_dataset, "--supervised", "-o",
            "trust", "belief", "accuracy"
        )
        results = yaml.safe_load(capsys.readouterr().out)["voting"]
        assert results["trust"] == {0: 1, 1: 1, 2: 1}
        assert results["belief"] == {
            0: {2: 1, 3: 1},
            1: {1: 1, 2: 1},
            2: {1: 1},
            3: {2: 1, 3: 1}
        }
        # accuracy is not deterministic when there are ties for most-believed
        # values
        assert results["accuracy"] in (1 / 3, 2 / 3, 0)

    def test_accuracy_not_supervised(self, csv_dataset, capsys):
        with pytest.raises(SystemExit):
            self.run(
                "run", "-a", "voting", "-f", csv_dataset, "-o", "accuracy"
            )
        err_msg = "cannot calculate accuracy without --supervised"
        assert err_msg in capsys.readouterr().err

    def test_accuracy_undefined(self, capsys, tmpdir):
        ds = tmpdir.join("newdata.csv")
        ds.write("\n".join((
            "1,2,3,4",
            "1,2,3,4",
        )))
        self.run("run", "-a", "voting", "-f", str(ds), "-s", "-o", "accuracy")
        results = yaml.safe_load(capsys.readouterr().out)["voting"]
        assert results["accuracy"] is None

    def test_graph_generation(self, client, csv_dataset, capsys, tmpdir):
        outfile = tmpdir.join("mypic.png")
        self.run("graph", "-f", csv_dataset, "-o", str(outfile))
        with open(str(outfile), "rb") as f:
            assert is_valid_png(f)

        # Missing dataset
        with pytest.raises(SystemExit):
            self.run("graph", "-o", str(outfile))
        assert "--dataset" in capsys.readouterr().err

        # Missing output file
        with pytest.raises(SystemExit):
            self.run("graph", "-f", csv_dataset)
        assert "--outfile" in capsys.readouterr().err

        # Renderer options
        rend1 = client.get_graph_renderer(
            self.get_parsed_args(
                "graph", "-f", csv_dataset, "-o", str(outfile), "--width",
                "200", "--node-radius", "25", "--font-size", "14", "--spacing",
                "4", "--line-width", "9", "--node-border-width", "34",
                "--one-indexed"
            )
        )
        assert rend1.width == 200
        assert rend1.node_radius == 25
        assert rend1.spacing == 4
        assert rend1.font_size == 14
        assert rend1.line_width == 9
        assert rend1.node_border_width == 34
        assert not rend1.zero_indexed

        # Check that if we only give some options, the rest are kept at the
        # defaults
        rend2 = client.get_graph_renderer(
            self.get_parsed_args(
                "graph", "-f", csv_dataset, "-o", str(outfile), "--width",
                "200"
            )
        )
        assert rend2.width == 200
        assert rend2.node_radius == 50
        assert rend2.zero_indexed


class TestWebClient(ClientTestsBase):
    @pytest.fixture
    def test_client(self):
        return get_flask_app().test_client()

    def test_routing(self):
        class ExampleClass:
            def __init__(self, x):
                self.x = x

            @route("/some-test-url/<name>/<int:age>", methods=["POST"])
            def my_view_function(self, name, age):
                return (
                    "name is {}, age is {} and x is {}"
                    .format(name, age, self.x)
                )

        app = Flask("test_routing")
        route.add_routes(app, ExampleClass(42))
        client = app.test_client()

        resp1 = client.post("/some-test-url/joe/22")
        assert resp1.status_code == 200
        assert resp1.data == b"name is joe, age is 22 and x is 42"

        # GET should not be allowed
        resp2 = client.get("/some-test-url/jim/91")
        assert resp2.status_code == 405

    def test_home(self, test_client):
        resp = test_client.get("/")
        html = resp.data.decode()
        assert "<body>" in html
        assert "truth discovery tool" in html.lower()
        assert '&#34;average_log&#34;: &#34;AverageLog&#34;' in html

    def test_run_fail(self, test_client):
        # Missing parameters
        resp1 = test_client.get("/run/")
        assert resp1.status_code == 400
        err_msg = "'algorithm' and 'matrix' parameters are required"
        assert resp1.json == {"ok": False, "error": err_msg}
        resp2 = test_client.get("/run/", query_string={"algorithm": "sums"})
        assert resp2.status_code == 400
        assert resp2.json == {"ok": False, "error": err_msg}

        # Invalid algorithm
        data = {"algorithm": "hmm", "matrix": "1,2,3"}
        resp3 = test_client.get("/run/", query_string=data)
        assert resp3.status_code == 400
        assert not resp3.json["ok"]
        assert "invalid algorithm label 'hmm'" in resp3.json["error"]

        # Invalid matrices
        invalid_matrices = (
            "hello",
            "1,2,3\n1,2,3,4",     # inconsistent shape
            "1,2,3ree\n1,2,3,4",  # non-float values
        )
        for matrix in invalid_matrices:
            data = {"algorithm": "sums", "matrix": matrix}
            resp = test_client.get("/run/", query_string=data)
            assert resp.status_code == 400
            assert not resp.json["ok"]
            assert "invalid matrix CSV" in resp.json["error"]

        # Invalid parameters
        invalid_params = (
            ("random-string", "parameters must be in the form 'key=value'"),
            ("priors=blah", "'blah' is not a valid PriorBelief"),
            ("priors=voted\nrandom-string", "in the form 'key=value'"),
        )
        for params, exp_err in invalid_params:
            data = {
                "algorithm": "sums",
                "matrix": "1,2,3\n4,5,6",
                "parameters": params
            }
            resp = test_client.get("/run/", query_string=data)
            assert resp.status_code == 400
            assert not resp.json["ok"]
            assert exp_err in resp.json["error"]

    def test_run_success(self, test_client):
        data = {
            "algorithm": "investment",
            # leading and trailing whitespace shouldn't matter
            "matrix": " 1,2,3,\n_,2,3,4   \n\n\n ",
            "parameters": "g=1.2\niterator=fixed-13"
        }
        resp = test_client.get("/run/", query_string=data)
        assert resp.status_code == 200
        assert resp.json["ok"]
        output = resp.json["data"]
        assert set(output.keys()) == {"investment"}
        assert output["investment"]["iterations"] == 13

    def test_run_multiple_algorithms(self, test_client, dataset):
        algorithms = ["investment", "sums", "voting"]
        data = {
            "algorithm": algorithms,
            "matrix": dataset.to_csv(),
            # Check that parameters apply to all algorithms, and are ignored if
            # not applicable
            "parameters": "iterator=fixed-11\ng=1.16"
        }
        resp = test_client.get("/run/", query_string=data)
        assert resp.status_code == 200
        assert resp.json["ok"]
        output = resp.json["data"]
        assert set(output.keys()) == set(algorithms)
        assert output["investment"]["iterations"] == 11
        assert output["sums"]["iterations"] == 11
        assert output["voting"]["iterations"] is None

        assert output["voting"]["trust"] == {str(i): 1 for i in range(4)}
        assert output["sums"]["trust"] == {
            "0": 1.0, "1": 0.5323216608830517, "2": 0.3472167514190226,
            "3": 5.9742483641484065e-05
        }
        assert output["investment"]["trust"] == {
            "0": 1.0, "1": 0.8725062856135304, "2": 0.6845094785538429,
            "3": 0.1211362221507375
        }

        assert "messages" in resp.json
        assert set(resp.json["messages"]) == {
            "Ignored parameters for 'MajorityVoting': g, iterator",
            "Ignored parameters for 'Sums': g",
        }

    def test_results_diff(self, test_client):
        dataset1 = MatrixDataset(ma.masked_values([
            [1, 0],
            [0, 2]
        ], 0))
        dataset2 = MatrixDataset(ma.masked_values([
            [1, 2],
            [0, 2]
        ], 0))

        request_data1 = {
            "algorithm": "voting",
            "matrix": dataset1.to_csv()
        }
        resp1 = test_client.get("/run/", query_string=request_data1)
        assert resp1.status_code == 200

        request_data2 = {
            # Test diffs with multiple algorithms
            "algorithm": ["voting", "investment"],
            "matrix": dataset2.to_csv(),
            "previous_results": json.dumps(resp1.json["data"]["voting"])
        }
        resp2 = test_client.get("/run/", query_string=request_data2)
        assert resp2.status_code == 200
        vot_out = resp2.json["data"]["voting"]
        assert "diff" in vot_out
        assert vot_out["diff"]["trust"] == {"0": 0, "1": 0}  # no trust changes
        assert vot_out["diff"]["belief"] == {
            "0": {"1.0": -0.5}, "1": {"2.0": 0}
        }
        inv_out = resp2.json["data"]["investment"]
        assert "diff" in inv_out
        assert inv_out["diff"]["trust"] == {
            "0": -0.7731216864273125, "1": 0.0
        }
        assert inv_out["diff"]["belief"] == {
            "0": {"1.0": -0.9354768146506873}, "1": {"2.0": 0.0}
        }

    def test_results_diff_invalid_previous_results(self, test_client):
        # Invalid JSON for previous results
        data1 = {
            "algorithm": "voting",
            "matrix": "1,2,3,\n_,2,3,4",
            "previous_results": "invalid json!"
        }
        resp1 = test_client.get("/run/", query_string=data1)
        assert resp1.status_code == 400
        assert not resp1.json["ok"]
        exp_err_msg = "'previous_results' is invalid: invalid JSON"
        assert exp_err_msg in resp1.json["error"]

        # Keys missing in previous results JSON
        data2 = {
            "algorithm": "voting",
            "matrix": "1,2,3,\n_,2,3,4",
            # 'belief' and others are missing
            "previous_results": '{"trust": {"0": 1}}'
        }
        resp2 = test_client.get("/run/", query_string=data2)
        assert resp2.status_code == 400
        assert not resp2.json["ok"]
        exp_err_msg = ("'previous_results' is invalid: required field "
                       "'belief' missing")
        assert exp_err_msg in resp2.json["error"]

    def test_get_json_graph(self, test_client, dataset):
        data = {
            "matrix": dataset.to_csv(),
            "algorithm": "sums"
        }
        resp = test_client.get("/run/", query_string=data)
        assert resp.status_code == 200
        output = resp.json
        assert "data" in output
        assert "sums" in output["data"]
        assert "imagery" in output["data"]["sums"]
        assert "graph" in output["data"]["sums"]["imagery"]
        json_string = output["data"]["sums"]["imagery"]["graph"]
        # Check the image is valid JSON
        obj = json.loads(json_string)
        # Check object is as expected
        assert "width" in obj
        assert "height" in obj
        assert "entities" in obj
        assert isinstance(obj["entities"], list)

    def test_get_json_animated_gif(self, test_client, dataset):
        data = {
            "matrix": dataset.to_csv(),
            "algorithm": "sums",
            "parameters": "iterator=fixed-4"
        }
        resp = test_client.get("/run/", query_string=data)
        assert resp.status_code == 200
        output = resp.json
        assert "data" in output
        assert "sums" in output["data"]
        assert "imagery" in output["data"]["sums"]
        assert "animation" in output["data"]["sums"]["imagery"]
        json_string = output["data"]["sums"]["imagery"]["animation"]
        # Check animation is valid JSON
        obj = json.loads(json_string)
        # Check object is as expected
        assert "fps" in obj
        assert "frames" in obj
        assert isinstance(obj["frames"], list)
        assert len(obj["frames"]) == 5
        assert isinstance(obj["frames"][0], dict)
        assert "width" in obj["frames"][0]
        assert "height" in obj["frames"][0]
        assert "entities" in obj["frames"][0]

    def test_voting_imagery(self, test_client, dataset):
        """
        Only the graph should be present in imagery for MajorityVoting results,
        since animation is not applicable
        """
        data = {
            "matrix": dataset.to_csv(),
            "algorithm": "voting"
        }
        resp = test_client.get("/run/", query_string=data)
        assert resp.status_code == 200
        output = resp.json
        assert "data" in output
        assert "voting" in output["data"]
        assert "imagery" in output["data"]["voting"]
        assert "graph" in output["data"]["voting"]["imagery"]
        assert "animation" not in output["data"]["voting"]["imagery"]

    def test_empty_dataset(self, test_client):
        data = {
            "algorithm": "sums",
            "matrix": ",,,\n,,,\n,,,\n,,,"
        }
        resp = test_client.get("/run/", query_string=data)
        assert resp.status_code == 400
        err_msg = "Cannot run algorithm on empty dataset"
        assert resp.json == {"ok": False, "error": err_msg}

    def test_non_convergence(self, test_client):
        data = {
            "algorithm": "sums",
            "matrix": "1,2,3\n4,5,6",
            "parameters": "iterator=l1-convergence-0-limit-25"
        }
        resp = test_client.get("/run/", query_string=data)
        assert resp.status_code == 500
        err_msg = "Did not converge in 25 iterations"
        assert resp.json == {"ok": False, "error": err_msg}

    def test_too_many_iterations(self, test_client):
        data = {
            "algorithm": "sums",
            "matrix": "1,2,3\n4,5,6",
            "parameters": "iterator=fixed-1000"
        }
        resp = test_client.get("/run/", query_string=data)
        assert resp.status_code == 400
        err_msg = "Cannot perform more than 200 iterations"
        assert resp.json == {"ok": False, "error": err_msg}



class BaseBackend:
    """
    Class to represent a method of drawing a list of :any:`Entity` objects to
    a file
    """
    def draw_entities(self, entities, outfile, width, height):
        """
        :param entities: an iterable of :any:`Entity` objects
        :param outfile:  file object to write to
        :param width:    width of image in px
        :param height:   height of image in px
        """
        raise NotImplementedError("Must be implemented in child classes")



import numpy as np

from truthdiscovery.algorithm.base import BaseIterativeAlgorithm


class AverageLog(BaseIterativeAlgorithm):
    """
    AverageLog by Pasternack and Roth.

    Similar to Sums (and uses the same belief update step), but updates source
    trust as average claim belief weighted by log(number of claims).
    """
    def _run(self, data):
        trust = np.zeros((data.num_sources,))
        belief = self.get_prior_beliefs(data)
        self.log(data, trust, belief)

        # Pre-compute the number of claims made by each source and log
        # weighting, since this is used in each iteration and does not change.
        # Note that the number of claims made by s_i is the sum of the i-th row
        # of the claims matrix, so we multiply by [1 ... 1].T to get the counts
        # for all sources in one operation
        claim_counts = data.sc @ np.ones((data.num_claims,))

        weights = np.log(claim_counts) / claim_counts

        while not self.iterator.finished():
            # Entry-wise multiplication
            new_trust = weights * (data.sc @  belief)
            belief = data.sc.T @ new_trust

            # Normalise as with sums
            new_trust = new_trust / max(new_trust)
            belief = belief / max(belief)

            self.iterator.compare(new_trust, trust)
            trust = new_trust
            self.log(data, trust, belief)

        return trust, belief



import numpy as np

from truthdiscovery.algorithm.base import BaseIterativeAlgorithm, PriorBelief
from truthdiscovery.exceptions import EarlyFinishError


class Investment(BaseIterativeAlgorithm):
    """
    Investment by Pasternack and Roth.

    Sources invest uniformly in their claims and receive returns; the source
    who invested most in a claim receives the most return from that claim
    """
    priors = PriorBelief.VOTED
    g = 1.2

    def __init__(self, *args, g=None, **kwargs):
        """
        :param g: exponent in non-linear function used to grow claim beliefs
                  at each iteration (optional)
        """
        if g is not None:
            self.g = g
        super().__init__(*args, **kwargs)

    def update_trust(self, old_trust, claim_counts, sc_mat, belief):
        """
        :return: an updated trust vector
        """
        # The amount each source has to invest in its claims
        investment_amounts = old_trust / claim_counts
        # The amount each claim receives in investment from its sources
        claim_investments = sc_mat.T @ investment_amounts
        if np.any(claim_investments == 0):
            raise EarlyFinishError(
                "Investment in at least one claim has become zero"
            )
        # Trust update can be expressed as the of entry-wise product of
        # investment amounts and the following matrix, which is obtained
        # by column-wise division (each column in sc is divided by the
        # corresponding entry in claim_investments)

        # (Note: using '/' here will result in a dense numpy array: we use
        # multiply() to get a sparse result instead)
        mat = sc_mat.multiply(1 / claim_investments)
        return investment_amounts * (mat @ belief)

    def _run(self, data):
        claim_counts = data.sc @ np.ones((data.num_claims,))
        trust = np.ones((data.num_sources,))
        belief = self.get_prior_beliefs(data)
        self.log(data, trust, belief)

        while not self.iterator.finished():
            try:
                new_trust = self.update_trust(
                    trust, claim_counts, data.sc, belief
                )
            except EarlyFinishError:
                break
            belief = (data.sc.T @ (new_trust / claim_counts)) ** self.g

            new_trust = new_trust / max(new_trust)
            belief = belief / max(belief)

            self.iterator.compare(new_trust, trust)
            trust = new_trust
            self.log(data, trust, belief)

        return trust, belief



import numpy as np

from truthdiscovery.algorithm.base import PriorBelief
from truthdiscovery.exceptions import EarlyFinishError
from truthdiscovery.algorithm.investment import Investment
from truthdiscovery.utils.iterator import FixedIterator


class PooledInvestment(Investment):
    """
    PooledInvestment by Pasternack and Roth.

    Refinement of Investment that makes use of mutual exclusion amongst claims
    """
    g = 1.4
    priors = PriorBelief.UNIFORM

    def get_default_iterator(self):
        """
        Override the default iterator to perform only 10 iterations
        """
        return FixedIterator(10)

    def _run(self, data):
        claim_counts = data.sc @ np.ones((data.num_claims,))
        trust = np.ones((data.num_sources,))
        belief = self.get_prior_beliefs(data)
        self.log(data, trust, belief)

        while not self.iterator.finished():
            # Trust update is the same as for Investment
            try:
                new_trust = self.update_trust(
                    trust, claim_counts, data.sc, belief
                )
            except EarlyFinishError:  # pragma: no cover
                break
            # 'Invest' trust in claims, grow with non-linear function, and
            # update belief
            base_returns = data.sc.T @ (new_trust / claim_counts)
            returns = base_returns ** self.g
            belief = base_returns * (returns / (data.mut_ex @ returns))

            new_trust = new_trust / max(new_trust)
            belief = belief / max(belief)

            self.iterator.compare(new_trust, trust)
            trust = new_trust
            self.log(data, trust, belief)

        return trust, belief



from io import BytesIO, StringIO
import json

import imageio

from truthdiscovery.graphs.backends import JsonBackend, PngBackend
from truthdiscovery.graphs.colours import ResultsGradientColourScheme
from truthdiscovery.graphs.draw import GraphRenderer


class BaseAnimator:
    """
    Base class to handle creating an animation of results-coloured graphs over
    the course of an algorithm's iteration.
    """
    # Must be defined in child classes
    buffer_cls = None
    supported_backends = ()

    def __init__(self, renderer=None, frame_duration=0.2):
        """
        :param renderer:       :any:`GraphRenderer` (or sub-class) object, or
                               None to use a default renderer
        :param frame_duration: duration in seconds for each frame
        """
        if renderer is None:
            default_backend_cls = self.supported_backends[0]
            renderer = GraphRenderer(backend=default_backend_cls())
        self.renderer = renderer
        self.fps = 1 / frame_duration

        if not isinstance(self.renderer.backend, self.supported_backends):
            clses = ", ".join(cls.__name__ for cls in self.supported_backends)
            raise TypeError(
                "Invalid renderer backend type {}, must be one of: {}"
                .format(self.renderer.backend.__class__, clses)
            )

    def animate(self, outfile, algorithm, dataset, show_progress=True):
        """
        :param outfile:       file object to write to
        :param algorithm:     :any:`BaseIterativeAlgorithm` sub-class instance
        :param dataset:       :any:`Dataset` object
        :param show_progress: if True, show a progress bar in the animation
        """
        frames = self.get_frames(algorithm, dataset, show_progress)
        self.write(frames, outfile)

    def get_frames(self, algorithm, dataset, show_progress):
        """
        A generator of ``buffer_cls`` objects for each frame in the animation
        """
        # Note: must collect all results so we can get total number of
        # iterations to work out completion percentage at each step
        all_results = tuple(algorithm.run_iter(dataset))
        num_iterations = len(all_results) - 1

        for i, results in enumerate(all_results):
            self.renderer.colours = ResultsGradientColourScheme(results)
            # Draw frame to in-memory buffer
            buf = self.buffer_cls()
            progress = i / num_iterations if show_progress else None
            self.renderer.render(dataset, buf, animation_progress=progress)
            buf.seek(0)
            yield buf

    def write(self, frames, outfile):
        """
        :param frames:  iterable of frames as ``buffer_cls`` instances
        :param outfile: file object to write to
        """
        raise NotImplementedError("Must be implemented in child classes")


class GifAnimator(BaseAnimator):
    """
    Create an animation as a GIF. Must be used with a renderer with PNG backend
    """
    buffer_cls = BytesIO
    supported_backends = (PngBackend,)

    def write(self, frames, outfile):
        kwargs = dict(format="gif", mode="I", fps=self.fps)
        with imageio.get_writer(outfile, **kwargs) as writer:
            for buf in frames:
                writer.append_data(imageio.imread(buf))


class JsonAnimator(BaseAnimator):
    """
    Create a JSON object containing each frame of the animation
    """
    buffer_cls = StringIO
    supported_backends = (JsonBackend,)

    def write(self, frames, outfile):
        obj = {"fps": self.fps, "frames": [json.load(buf) for buf in frames]}
        json.dump(obj, outfile)



import os

import numpy as np
import numpy.ma as ma

from truthdiscovery.input.dataset import Dataset


def csv_to_masked_array(fileobj):
    """
    Parse a CSV file and return a numpy masked array

    :param fileobj:     fileobj to read from
    :return:            a numpy masked array representing the matrix encoded by
                        the CSV
    :raises ValueError: if CSV contains values that cannot be converted to
                        floats, or if shape is invalid
    """
    entries = []
    width = None
    csv = fileobj.read().strip()
    for i, line in enumerate(csv.split(os.linesep)):
        # Split by comma and remove extra whitespace from each value
        row = list(map(str.strip, line.split(",")))
        if width is None:
            width = len(row)
        elif len(row) != width:
            raise ValueError("Expected {} entries in row {}, got {}"
                             .format(width, i + 1, len(row)))
        entries.append([float(val) if val else np.nan for val in row])

    matrix = np.array(entries)
    return ma.masked_array(matrix, np.isnan(matrix))


class MatrixDataset(Dataset):
    """
    A truth discovery dataset given as a matrix, where columns correspond to
    variables, and rows correspond to sources.

    Let ``X_1, ..., X_n`` be the variables, and ``s_1, ... s_m`` be the
    sources. A value ``v`` at entry ``(i, j)`` means that ``s_i`` asserts that
    ``X_j = v``.

    No entry at ``(i, j)`` means that ``s_i`` does not make any assertions
    regarding the value of ``X_j``.
    """
    def __init__(self, sv_mat, *args, **kwargs):
        """
        :param sv_mat: source-variables matrix as a 2D numpy array. May be a
                       masked array to encode missing values
        :raises ValueError: if the dimension of the input is invalid
        """
        self.sv = sv_mat
        if self.sv.ndim != 2:
            raise ValueError("Source/variables matrix must be two dimensional")

        super().__init__(self.get_triples(), *args, **kwargs)

    def get_triples(self):
        """
        :yield: triples ``(source, var, val)`` for each non-empty entry in the
                matrix. Source and variable labels are defined as their row and
                column numbers respectively.
        """
        for source, row in enumerate(self.sv):
            for var, val in enumerate(row):
                if not ma.is_masked(val):
                    yield (source, var, val)

    @classmethod
    def from_csv(cls, fileobj):
        """
        Load a matrix from a CSV file

        :param fileobj:     file object to read from
        :return:            a :any:`MatrixDataset` object
        :raises ValueError: if CSV is invalid
        """
        try:
            return cls(csv_to_masked_array(fileobj))
        except ValueError as ex:
            raise ValueError("invalid matrix CSV: {}".format(ex))

    def to_csv(self):
        """
        :return: a string representation of the dataset in CSV format
        """
        return "\n".join(
            ",".join("" if ma.is_masked(val) else str(val) for val in row)
            for row in self.sv
        )



import os.path

import numpy.ma as ma

from truthdiscovery.algorithm import Sums
from truthdiscovery.input import MatrixDataset
from truthdiscovery.graphs import (
    MatrixDatasetGraphRenderer,
    ResultsGradientColourScheme
)
from truthdiscovery.utils import ConvergenceIterator, DistanceMeasures


if __name__ == "__main__":
    ds1 = MatrixDataset(ma.masked_values([
        [1, 1],
        [1, 9],
        [9, 1],
        [2, 2],
    ], 9))
    ds2 = MatrixDataset(ma.masked_values([
        [1, 1, 9, 9, 9],
        [1, 9, 9, 9, 9],
        [9, 1, 9, 9, 9],
        [2, 2, 9, 9, 9],
        [9, 9, 1, 1, 1],
        [9, 9, 1, 1, 1],
        [9, 9, 2, 2, 2],
    ], 9))

    alg = Sums(iterator=ConvergenceIterator(DistanceMeasures.L2, 0.0000001))
    res1 = alg.run(ds1)
    res2 = alg.run(ds2)

    scenarios = [(res1, ds1, "results1.png"), (res2, ds2, "results2.png")]
    for res, ds, name in scenarios:
        path = os.path.join("/tmp", name)
        rend = MatrixDatasetGraphRenderer(
            colours=ResultsGradientColourScheme(res)
        )
        with open(path, "wb") as outfile:
            rend.render(ds, outfile)

    for source, trust2 in res2.trust.items():
        trust1 = res1.trust.get(source, 9)
        print("{}: {:.4f}\t{:.4f}".format(source, trust1, trust2))



import random

import numpy.ma as ma

from truthdiscovery.input.matrix_dataset import (
    csv_to_masked_array,
    MatrixDataset
)


class SupervisedData:
    """
    A class to store a dataset for which the true values of a subset of the
    variables is known
    """
    def __init__(self, dataset, true_values):
        """
        :param dataset:     a :any:`Dataset` (or sub-class) object
        :param true_values: dict of the form ``{var_label: true_value, ...}``
                            pairs for the known true values
        """
        self.data = dataset
        self.values = true_values

    def get_accuracy(self, results):
        """
        Calculate the accuracy of truth discovery results, computed as the
        frequency of cases where the most believed value for a variable is the
        correct one, ignoring cases where only one value for a variable is
        claimed across all sources (in this case all algorithms will predict
        the same value).

        :param results: a :any:`Result` object
        :return: accuracy as a number in [0, 1]: 1 is best accuracy, 0 is worst
        :raises ValueError: if no true values are known, or if all variables
                            have only one claimed value
        """
        if not self.values:
            raise ValueError("No known true values")
        total = 0
        count = 0

        for var_label, true_value in self.values.items():
            # Skip if there is only one claimed value
            try:
                if len(results.belief[var_label]) == 1:
                    continue
            except KeyError:
                continue

            total += 1
            # Note: select value randomly if more than one most-believed value
            # exists
            most_believed = random.choice(
                list(results.get_most_believed_values(var_label))
            )
            if most_believed == true_value:
                count += 1
        if total == 0:
            raise ValueError(
                "No known variables where more than one claimed value exists"
            )
        return count / total

    @classmethod
    def from_csv(cls, fileobj):
        """
        Load a matrix from a CSV file along with true values. The format is the
        same as for loading an unsupervised matrix dataset, but the first row
        contains the true values.

        :param fileobj: file object to read from
        :return:        a :any:`SupervisedData` object representing the matrix
                        encoded by the CSV
        """
        # Load the whole thing as a matrix
        temp = csv_to_masked_array(fileobj)
        # Get true values from first row
        true_values = {i: v for i, v in enumerate(temp[0, :])
                       if not ma.is_masked(v)}
        sv_mat = temp[1:, :]
        return cls(MatrixDataset(sv_mat), true_values)



class ResultDiff:
    """
    Object to represent the difference between two :any:`Result` objects.
    Has attributes for differences in source trusts, claim beliefs, time taken
    and number of iterations
    """
    def __init__(self, res1, res2):
        self.trust = {}
        self.belief = {}
        self.time_taken = res2.time_taken - res1.time_taken
        self.iterations = None

        sources1 = res1.trust.keys()
        sources2 = res2.trust.keys()
        common_sources = set(sources1).intersection(sources2)
        for source in common_sources:
            self.trust[source] = res2.trust[source] - res1.trust[source]

        vars1 = res1.belief.keys()
        vars2 = res2.belief.keys()
        common_vars = set(vars1).intersection(vars2)
        for var in common_vars:
            diffs = {}
            vals1 = set(res1.belief[var].keys())
            vals2 = set(res2.belief[var].keys())
            common_vals = vals1.intersection(vals2)
            for val in common_vals:
                diffs[val] = res2.belief[var][val] - res1.belief[var][val]
            if diffs:
                self.belief[var] = diffs

        if res1.iterations is not None and res2.iterations is not None:
            self.iterations = res2.iterations - res1.iterations



from enum import Enum


class NodeType(Enum):
    SOURCE = "source"
    CLAIM = "claim"
    VARIABLE = "variable"


class GraphColourScheme:
    """
    Default colour scheme for graphs
    """
    def get_node_colour(self, node_type, hints):
        """
        :param node_type: value from :any:`NodeType` enumeration
        :param hints:     for variables and sources: var/source ID. For claims,
                          a tuple (var, val)
        :return: `(node_colour, label_colour, border_colour)`
        """
        node_colour = None
        if node_type == NodeType.SOURCE:
            node_colour = (0.133, 0.40, 0.40)  # turquoise
        elif node_type == NodeType.CLAIM:
            node_colour = (0.478, 0.624, 0.208)  # green
        else:
            node_colour = (0.667, 0.224, 0.224)  # red

        label_colour = (1, 1, 1)            # white labels
        border_colour = (0.15, 0.15, 0.15)  # dark grey border

        return node_colour, label_colour, border_colour

    def get_background_colour(self):
        """
        Return background colour for the graph
        """
        return (1, 1, 1)  # white

    def get_edge_colour(self):
        """
        Return colour for the edges between nodes
        """
        return (0.5, 0.5, 0.5)  # grey

    def get_animation_progress_colour(self):
        """
        Return colour for the bar displaying animation progress
        """
        return (0, 0, 0)  # black


class ResultsGradientColourScheme(GraphColourScheme):
    """
    Colour scheme where the colour for a source/claim depends on its
    trust/belief score in a given set of results
    """
    # Colours obtained from here:
    # http://colorbrewer2.org/?type=sequential&scheme=Greens&n=9
    COLOURS = [
        # Lowest value (pale colour)
        (0.9686274509803922, 0.9882352941176471, 0.9607843137254902),
        (0.8980392156862745, 0.9607843137254902, 0.8784313725490196),
        (0.7803921568627451, 0.9137254901960784, 0.7529411764705882),
        (0.6313725490196078, 0.8509803921568627, 0.6078431372549019),
        (0.4549019607843137, 0.7686274509803922, 0.4627450980392157),
        (0.2549019607843137, 0.6705882352941176, 0.36470588235294116),
        (0.13725490196078433, 0.5450980392156862, 0.27058823529411763),
        (0.0, 0.42745098039215684, 0.17254901960784313),
        # Highest value (dark green)
        (0.0, 0.26666666666666666, 0.10588235294117647)
    ]
    LIGHT_LABEL = (1, 1, 1)  # white
    DARK_LABEL = (0, 0, 0)   # black

    def __init__(self, results):
        self.results = results

    def get_graded_colours(self, level):
        """
        :param level: a float in [0, 1]
        :return: (interior, label) RGB triples for the colours corresponding to
                 the given level
        """
        # Quantise level to an integer in [0, num_colours)
        index = min(int(level * len(self.COLOURS)), len(self.COLOURS) - 1)
        label_colour = self.DARK_LABEL if level < 0.5 else self.LIGHT_LABEL
        return self.COLOURS[index], label_colour

    def get_node_colour(self, node_type, hints):
        defaults = super().get_node_colour(node_type, hints)
        default_node, default_label, default_border = defaults

        # We can only use graded colours for sources and claims, so stick to
        # defaults in parent class for variables
        if node_type == NodeType.VARIABLE:
            return defaults

        # Get level from hint
        if isinstance(hints, tuple):
            var, val = hints
            level = self.results.belief[var][val]
        else:
            source = hints
            level = self.results.trust[source]

        node_colour, label_colour = self.get_graded_colours(level)
        return node_colour, label_colour, default_border


class PlainColourScheme(GraphColourScheme):
    """
    Plain black and white colour scheme with transparent background
    """
    def get_background_colour(self):
        return (1, 1, 1, 0)

    def get_node_colour(self, *args):
        return (1, 1, 1), (0, 0, 0), (0, 0, 0)

    def get_edge_colour(self):
        return (0, 0, 0)



from enum import Enum
import inspect
import time

import numpy as np

from truthdiscovery.exceptions import EmptyDatasetError
from truthdiscovery.output import Result
from truthdiscovery.utils.iterator import FixedIterator


class PriorBelief(Enum):
    """
    Enumeration of possible choices for prior beliefs, used in iterative
    algorithms

    TODO: need to write a brief explanation of each choice
    """
    FIXED = "fixed"
    VOTED = "voted"
    UNIFORM = "uniform"


class BaseAlgorithm:
    """
    Base class for truth discovery algorithms
    """
    def run(self, data):
        """
        Run the algorithm on the given data

        :param data: input data as a :any:`Dataset` object
        :return: the results as a :any:`Result` tuple

        :raises EmptyDatasetError: if the dataset contains no claims
        """
        if data.num_claims == 0:
            raise EmptyDatasetError("Cannot run algorithm on empty dataset")
        # Real work must be performed in child classes

    @classmethod
    def get_parameter_names(cls):
        """
        Find the names of all parameters that this algorithm make take in its
        constructor (including parameters from parent classes)

        :return: a set of parameter names as strings
        """
        def yield_names(alg_cls):
            for param in inspect.signature(alg_cls.__init__).parameters:
                if param not in ("self", "args", "kwargs"):
                    yield param
            for base in alg_cls.__bases__:
                yield from yield_names(base)
        return set(yield_names(cls))


class BaseIterativeAlgorithm(BaseAlgorithm):
    """
    Base class for functionality common to algorithms that iteratively compute
    trust and belief
    """
    iterator = None
    priors = PriorBelief.FIXED
    results_log = None

    def __init__(self, iterator=None, priors=None):
        """
        :param iterator: :any:`Iterator` object to control when iteration stops
                         (optional)
        :param priors:   value from :any:`PriorBelief` enumeration to specify
                         which prior belief values are used (optional)
        """
        self.iterator = iterator or self.get_default_iterator()
        if priors is not None:
            self.priors = priors

    def get_default_iterator(self):
        """
        Return the :any:`Iterator` object to use by default, if the user does
        not provide one

        :return: an :any:`Iterator` object
        """
        return FixedIterator(20)

    def get_prior_beliefs(self, data):
        """
        :param data:        input data as a :any:`Dataset` object
        :return:            a numpy array of prior belief values for claims, in
                            claim ID order
        :raises ValueError: if ``self.prior`` is not an item from the
                            :any:`PriorBelief` enumeration
        """
        if self.priors == PriorBelief.FIXED:
            return np.full((data.num_claims,), 0.5)

        if self.priors == PriorBelief.VOTED:
            source_counts = data.sc.T @ np.ones((data.num_sources,))
            return source_counts / (data.mut_ex @ source_counts)

        if self.priors == PriorBelief.UNIFORM:
            return 1 / (data.mut_ex @ np.ones((data.num_claims,)))

        raise ValueError(
            "Invalid prior belief type: '{}'".format(self.priors)
        )

    def run(self, data):
        super().run(data)
        self.iterator.reset()
        self.start_time = time.time()
        self.results_log = None
        trust, belief = self._run(data)
        end_time = time.time()
        return Result(
            trust=data.get_source_trust_dict(trust),
            belief=data.get_belief_dict(belief),
            time_taken=end_time - self.start_time,
            iterations=self.iterator.it_count
        )

    def run_iter(self, data):
        """
        Return a generator of partial :any:`Result` objects as the algorithm
        iterates
        """
        super().run(data)
        self.iterator.reset()
        self.start_time = time.time()
        self.results_log = []
        _t, _b = self._run(data)
        yield from self.results_log

    def _run(self, data):
        """
        Internal method for running the algorithm, to avoid including
        boilerplate code in each subclass

        :param data: :any:`Dataset` object
        :return: a tuple ``(trust, belief)``, where ``trust`` is a numpy
                 array of source trusts, and ``belief`` is a numpy array of
                 claim beliefs, both ordered as in the input data
        """
        raise NotImplementedError("Must be implemented in child classes")

    def log(self, data, trust, belief):
        """
        If logging is enabled, append the given trust and belief scores to the
        log
        """
        if self.results_log is not None:
            res = Result(
                trust=data.get_source_trust_dict(trust),
                belief=data.get_belief_dict(belief),
                time_taken=time.time() - self.start_time,
                iterations=self.iterator.it_count
            )
            self.results_log.append(res)



"""
Script to demonstrate running an algorithm on some small made-up data
"""
import numpy.ma as ma

from truthdiscovery.algorithm import (
    AverageLog,
    Investment,
    MajorityVoting,
    PooledInvestment,
    Sums,
    TruthFinder
)
from truthdiscovery.input import MatrixDataset


def main():
    """
    Run an algorithm and print results
    """
    data = MatrixDataset(ma.masked_values([
        [1, 4, 0, 5, 7, 0],
        [0, 2, 0, 4, 9, 0],
        [1, 0, 2, 3, 8, 4],
        [4, 0, 2, 5, 0, 0],
        [4, 0, 3, 3, 4, 4]
    ], 0))

    alg = AverageLog()

    results = alg.run(data)
    print("trust scores:")
    for source, trust_val in results.trust.items():
        print("  source {}: {:.3f}".format(source, trust_val))
    print("")
    print("belief scores:")
    for var in sorted(results.belief):
        print("  variable {}:".format(var))
        beliefs = results.belief[var]
        for val in sorted(beliefs):
            print("    {}: {:.3f}".format(val, beliefs[val]))


if __name__ == "__main__":
    main()



import numpy as np
import numpy.ma as ma
import pytest

from truthdiscovery.algorithm import MajorityVoting
from truthdiscovery.input import (
    Dataset,
    FileDataset,
    FileSupervisedData,
    IDMapping,
    MatrixDataset,
    SupervisedData,
    SyntheticData
)
from truthdiscovery.output import Result


class TestDataset:
    @pytest.fixture
    def data(self):
        triples = (
            ("john", "wind", "very windy"),
            ("paul", "wind", "not very windy"),
            ("george", "wind", "very windy"),
            ("ringo", "wind", "not very windy at all"),

            ("john", "rain", "dry"),
            ("george", "rain", "wet"),

            ("john", "water", "wet"),  # re-use value
            ("paul", "water", "drink"),
            ("george", "water", "drink"),

            # Mix up the order of variables
            ("ringo", "rain", "dry"),
        )
        return Dataset(triples)

    def test_num_sources_variables_claims(self, data):
        assert data.num_sources == 4
        assert data.num_variables == 3
        assert data.num_claims == 7

    def test_claims_matrix(self, data):
        expected_claim_mat = np.array([
            [1, 0, 0, 1, 0, 1, 0],
            [0, 1, 0, 0, 0, 0, 1],
            [1, 0, 0, 0, 1, 0, 1],
            [0, 0, 1, 1, 0, 0, 0],
        ])
        assert data.sc.shape == expected_claim_mat.shape
        assert np.array_equal(data.sc.toarray(), expected_claim_mat)

    def test_mut_ex_matrix(self, data):
        expected_mut_ex = np.array([
            [1, 1, 1, 0, 0, 0, 0],
            [1, 1, 1, 0, 0, 0, 0],
            [1, 1, 1, 0, 0, 0, 0],
            [0, 0, 0, 1, 1, 0, 0],
            [0, 0, 0, 1, 1, 0, 0],
            [0, 0, 0, 0, 0, 1, 1],
            [0, 0, 0, 0, 0, 1, 1],
        ])
        assert np.array_equal(data.mut_ex.toarray(), expected_mut_ex)

    def test_source_multiple_claims_for_a_single_variable(self):
        with pytest.raises(ValueError) as excinfo:
            Dataset((
                ("s1", "x", 4),
                ("s2", "x", 5),
                ("s1", "x", 5)
            ))

        err_msg = "Source 's1' claimed more than one value for variable 'x'"
        assert err_msg in str(excinfo.value)

        # Allow multiple claims; first one should be used
        data = Dataset((
            ("s1", "x", 5),
            ("s2", "x", 6),
            ("s3", "x", 5),
            ("s3", "x", 6),
        ), allow_multiple=True)
        exp_sc = np.array([
            [1, 0],
            [0, 1],
            [1, 0]
        ])
        assert np.array_equal(data.sc.toarray(), exp_sc)

    def test_num_connected_components(self):
        ds1 = Dataset([
            ("s1", "x", "a"),
            ("s2", "x", "b"),
            ("s3", "x", "a"),
        ])
        assert ds1.num_connected_components() == 1

        ds2 = Dataset([
            ("s1", "x", "a"),
            ("s2", "x", "b"),
            ("s3", "x", "a"),
            ("s4", "y", "a"),
            ("s5", "z", "a"),
            ("s6", "z", "a")
        ])
        assert ds2.num_connected_components() == 3


class TestIDMapping:
    def test_insert(self):
        mapping = IDMapping()
        # Should insert when querying for the first time
        assert mapping.get_id("hello") == 0
        assert mapping.get_id("goodbye") == 1
        # Should retrieve the same values when calling a second time
        assert mapping.get_id("hello") == 0
        # Labels can be any (hashable) type
        assert mapping.get_id(("this", "is", 4, "tuple")) == 2
        # Raise exception if disallowing inserts
        with pytest.raises(KeyError):
            mapping.get_id("something new", insert=False)
        # Should be bi-directional
        assert mapping.inverse[0] == "hello"
        assert mapping.inverse[1] == "goodbye"
        assert mapping.inverse[2] == ("this", "is", 4, "tuple")


class TestMatrixDataset:
    def test_create(self):
        # From a regular numpy array
        normal_arr = np.ones((2, 3))
        MatrixDataset(normal_arr)

        # From masked array
        masked_arr = ma.array(normal_arr,
                              mask=[[True, False, True], [False, True, False]])
        MatrixDataset(masked_arr)

    def test_num_sources_variables_claims(self):
        mat = MatrixDataset(np.array([
            [1, 4, 5],
            [2, 0, 5],
            [1, 1, 5],
            [3, 2, 5]
        ]))
        assert mat.num_sources == 4
        assert mat.num_variables == 3
        assert mat.num_claims == 8

    def test_dimension(self):
        arr = np.zeros((3, 3, 3))
        with pytest.raises(ValueError):
            MatrixDataset(arr)

    def test_from_csv(self, tmpdir):
        filepath = tmpdir.join("data.csv")
        filepath.write("\n".join([
            "1,,3, 2,6  ",  # extra whitespace should not matter
            ", 9,0,2,5",
            "3,9,  ,,1",
            "1,9  , 5.7,3,4",
            "5,1,3,1,1",
            "\n"  # new lines at the end of file should not matter
        ]))

        data = MatrixDataset.from_csv(filepath.open())
        expected_matrix = ma.masked_values([
            [1, 999, 3, 2, 6],
            [999, 9, 0, 2, 5],
            [3, 9, 999, 999, 1],
            [1, 9, 5.7, 3, 4],
            [5, 1, 3, 1, 1]
        ], 999)
        assert data.num_sources == 5
        assert data.num_variables == 5
        assert data.num_claims == 15
        assert np.array_equal(data.sv.mask, expected_matrix.mask)
        assert (data.sv == expected_matrix).all()

    def test_invalid_csv_shape(self, tmpdir):
        filepath = tmpdir.join("data.csv")
        filepath.write("\n".join([
            "1,2,",
            "1,2"
        ]))
        with pytest.raises(ValueError) as excinfo:
            MatrixDataset.from_csv(filepath.open())
        assert "Expected 3 entries in row 2, got 2" in str(excinfo.value)

    def test_from_csv_empty_rows(self, tmpdir):
        filepath = tmpdir.join("data.csv")
        filepath.write("\n".join([
            "1,2,",
            ",,",
            " ,\t,",
            "3,4,5"
        ]))
        data = MatrixDataset.from_csv(filepath.open())
        expected_matrix = ma.masked_values([
            [1, 2, 0],
            [0, 0, 0],
            [0, 0, 0],
            [3, 4, 5]
        ], 0)
        assert np.array_equal(data.sv.mask, expected_matrix.mask)
        assert (data.sv == expected_matrix).all()

    def test_from_csv_single_row_or_column(self, tmpdir):
        filepath1 = tmpdir.join("data1.csv")
        filepath1.write("1,,3,2,6")
        data1 = MatrixDataset.from_csv(filepath1.open())
        exp_sv1 = ma.masked_values([
            [1, 0, 3, 2, 6]
        ], 0)
        assert data1.num_sources == 1
        assert data1.num_variables == 4
        assert data1.num_claims == 4
        assert np.array_equal(data1.sv.mask, exp_sv1.mask)
        assert (data1.sv == exp_sv1).all()

        filepath2 = tmpdir.join("data2.csv")
        filepath2.write("1\n\n3\n2\n6")
        data2 = MatrixDataset.from_csv(filepath2.open())
        exp_sv2 = exp_sv1.T
        assert data2.num_sources == 4
        assert data2.num_variables == 1
        assert data2.num_claims == 4
        assert np.array_equal(data2.sv.mask, exp_sv2.mask)
        assert (data2.sv == exp_sv2).all()

    def test_claims_matrix(self):
        data = MatrixDataset(ma.masked_values([
            [7, 4, 7],
            [5, 1, -1],
            [-1, 2, 4],
            [7, -1, 2],
            [-1, 1, 2]
        ], -1))
        expected_claim_mat = np.array([
            [1, 1, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 1, 0],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [0, 0, 0, 0, 1, 0, 0, 1]
        ])
        assert data.sc.shape == expected_claim_mat.shape
        assert np.array_equal(data.sc.toarray(), expected_claim_mat)

    def test_mutual_exclusion_matrix(self):
        data = MatrixDataset(ma.masked_values([
            [7, 4, 7], [5, 1, -1], [-1, 2, 4], [7, -1, 2],
            [-1, 1, 2]
        ], -1))
        # Claims are:
        # 0: x=7
        # 1: y=4
        # 2: z=7
        # 3: x=5
        # 4: y=1
        # 5: y=2
        # 6: z=4
        # 7: z=2
        # Mutual exclusion groups are {0, 3}, {1, 4, 5}, {2, 6, 7}
        expected_mut_ex_mat = np.array([
            [1, 0, 0, 1, 0, 0, 0, 0],
            [0, 1, 0, 0, 1, 1, 0, 0],
            [0, 0, 1, 0, 0, 0, 1, 1],
            [1, 0, 0, 1, 0, 0, 0, 0],
            [0, 1, 0, 0, 1, 1, 0, 0],
            [0, 1, 0, 0, 1, 1, 0, 0],
            [0, 0, 1, 0, 0, 0, 1, 1],
            [0, 0, 1, 0, 0, 0, 1, 1],
        ])
        assert data.mut_ex.shape == expected_mut_ex_mat.shape
        assert np.array_equal(data.mut_ex.toarray(), expected_mut_ex_mat)

    def test_export_to_csv(self):
        data = MatrixDataset(ma.masked_values([
            # All full row
            [1, 2, 3, 4, 5, 6, 7, 8],
            # Mixed row
            [1, 2, 0, -123, 4, -2.3, 99.123, -123],
            # All empty row
            [-123, -123, -123, -123, -123, -123, -123, -123]
        ], -123))
        expected = "\n".join((
            "1.0,2.0,3.0,4.0,5.0,6.0,7.0,8.0",
            "1.0,2.0,0.0,,4.0,-2.3,99.123,",
            ",,,,,,,"
        ))
        assert data.to_csv() == expected


class TestSupervisedData:
    @pytest.fixture
    def dataset(self):
        return Dataset((
            ("s1", "x", 1), ("s1", "y", 2), ("s1", "z", 3), ("s1", "w", 4),
            ("s2", "x", 1), ("s2", "z", 3), ("s2", "w", 4),
            ("s3", "y", 2), ("s3", "z", 3), ("s3", "w", 4)
        ))

    def test_from_csv(self, tmpdir):
        filepath = tmpdir.join("data.csv")
        filepath.write("\n".join([
            "7,8,,100,",  # true values
            "1,,3,2,6",
            ",9,2,2,5",
            "3,9,,,1",
            "1,9,5,3,4",
            "5,1,3,1,1"
        ]))

        sup = SupervisedData.from_csv(filepath.open())
        data = sup.data
        expected_matrix = ma.masked_values([
            [1, 0, 3, 2, 6],
            [0, 9, 2, 2, 5],
            [3, 9, 0, 0, 1],
            [1, 9, 5, 3, 4],
            [5, 1, 3, 1, 1]
        ], 0)
        expected_values = {0: 7, 1: 8, 3: 100}
        assert data.num_sources == 5
        assert data.num_variables == 5
        assert data.num_claims == 15
        assert np.all(data.sv.mask == expected_matrix.mask)
        assert np.all(data.sv == expected_matrix)
        assert np.all(sup.values == expected_values)

    def test_accuracy(self, dataset):
        sup = SupervisedData(dataset, {"x": 5, "y": 6, "w": 8})
        test_data = (
            (2 / 3, {"x": {5: 1.0, 15: 0.8, -10: 0.5},           # correct
                     "y": {1: 0.99, 2: 0.8},                     # wrong
                     "z": {7: 0.5, 6: 0.7, -10: 0.999},          # unknown
                     "w": {1: 0.1, 2: 0.1, 3: 0.1, 8: 0.101}}),  # correct

            # Results where a variable has only one claimed value
            (1 / 2, {"x": {5: 1.0},                         # correct, only one
                     "y": {1: 0.99, 2: 0.8},                     # wrong
                     "z": {7: 0.5, 6: 0.7, -10: 0.999},          # unknown
                     "w": {1: 0.1, 2: 0.1, 3: 0.1, 8: 0.101}}),  # correct

            # Results where all claimed values are wrong
            (0, {"x": {4: 1, 2: 0.5}, "y": {4: 1}, "z": {4: 1}, "w": {4: 1}})
        )

        for expected_acc, belief in test_data:
            res = Result(
                trust={"s1": 1.5, "s2": 0.5, "s3": 0.5},
                belief=belief,
                time_taken=None
            )
            got_acc = sup.get_accuracy(res)
            assert got_acc == expected_acc

        # Results where there is a tie for most believed value
        var_beliefs = {
            "x": {5: 0.8, 5000: 0.8, 4: 0.6},  # tie: either correct or not
            "y": {6: 0.7, 1: 0.2},             # correct
            "z": {1: 0.5, 2: 0.4},             # unknown
            "w": {1: 0.5, 2: 0.4}              # wrong
        }
        res = Result(
            trust={0: 1.5, 1: 0.5, 2: 0.5}, belief=var_beliefs, time_taken=None
        )
        assert sup.get_accuracy(res) in (1 / 3, 2 / 3)

    def test_unknown_variable(self, dataset):
        sup = SupervisedData(dataset, {"hello": 42, "x": 4})
        res = Result(
            trust={"s1": 1.5, "s2": 0.5, "s3": 0.5},
            belief={
                "x": {4: 1, 2: 0.5},
                "y": {4: 1},
                "z": {4: 1},
                "w": {4: 1}
            },
            time_taken=None
        )
        # The variable "hello" should be ignored
        assert sup.get_accuracy(res) == 1

    def test_no_true_values_known(self, dataset):
        sup = SupervisedData(dataset, {})
        res = Result(
            trust={0: 0.5, 1: 0.5, 2: 0.5},
            belief={i: {4: 1} for i in range(4)},
            time_taken=None
        )
        with pytest.raises(ValueError):
            sup.get_accuracy(res)

    def test_not_enough_claimed_values(self):
        # If all variables have only one claimed value, the accuracy
        # calculation cannot be performed: an exception should be raised
        dataset = Dataset([
            ("s1", "x", 4),
            ("s2", "x", 4),
            ("s3", "x", 4),
            ("s1", "y", 1),
            ("s2", "y", 1),
            ("s3", "y", 1)
        ])
        sup = SupervisedData(dataset, {"x": 4, "y": 2})
        res = Result(
            trust={"s1": 0.3, "s2": 0.4, "s3": 0.4},
            belief={"x": {4: 1}, "y": {1: 1}},
            time_taken=None
        )
        with pytest.raises(ValueError):
            sup.get_accuracy(res)


class TestSyntheticData:
    """
    Test the SyntheticData class
    """
    def test_trust_as_list(self):
        SyntheticData(trust=[0, 1, 0.5])

    def test_invalid_trust_vector_shape(self):
        invalid_trust_scores = (
            np.array([]),
            np.array([
                [0.5, 0.5],
                [0.5, 0.5]
            ])
        )
        for trust in invalid_trust_scores:
            with pytest.raises(ValueError):
                SyntheticData(trust, 10)

    def test_trust_range_error(self):
        invalid_trust_scores = (
            np.array([-1, 0, 0]),
            np.array([2, 0, 0]),
            np.array([np.nan, 0, 0]),
            np.array([0, 1.1, 0])
        )
        for trust in invalid_trust_scores:
            with pytest.raises(ValueError):
                SyntheticData(trust, 10)

    def test_valid_trusts(self):
        valid_trust_scores = (
            np.array([0, 0, 0]),
            np.array([1, 1, 1]),
            np.array([1 / 3, 0.5, 0.4444444]),
        )
        for trust in valid_trust_scores:
            try:
                SyntheticData(trust, 10)
            except ValueError:  # pragma: no cover
                assert False, "Unexpected error for trust = {}".format(trust)

    def test_claim_probability(self):
        trust = np.full((5,), 0.5)
        synth = SyntheticData(trust, claim_probability=1)
        # If claims made with p=1 then all possible claims should be made
        assert (~synth.data.sv.mask).all()

    def test_invalid_claim_probability(self):
        invalid_probs = (0, -1, -0.5, -0.0000001, 1.0000001)
        trust = np.full((5,), 0.5)
        for prob in invalid_probs:
            with pytest.raises(ValueError):
                SyntheticData(trust, claim_probability=prob)

    def test_domain_size(self):
        invalid_domain_sizes = (-1, 0, 1)
        trust = np.full((5,), 0.5)
        for ds in invalid_domain_sizes:
            with pytest.raises(ValueError):
                SyntheticData(trust, domain_size=ds)

    def test_export_to_csv(self, tmpdir):
        synth = SyntheticData(np.array([0.5, 0.5]), num_variables=10)
        csv_string = synth.to_csv()
        lines = [line for line in csv_string.split("\n") if line]
        assert len(lines) == 3

        filename = tmpdir.join("mydata.csv")
        filename.write(csv_string)
        loaded = SupervisedData.from_csv(filename.open())
        assert np.array_equal(loaded.values, synth.values)
        assert np.array_equal(
            loaded.data.sc.toarray(),
            synth.data.sc.toarray()
        )


class TestImplications:
    @pytest.fixture
    def triples(self):
        return [
            ("s1", "x", 1),
            ("s1", "y", 2),
            ("s1", "z", 3),
            ("s1", "w", 4),

            ("s2", "x", 2),
            ("s2", "z", 4),
            ("s2", "w", 5),

            ("s3", "y", 3),
            ("s3", "z", 3),
            ("s3", "w", 5)
        ]

    def test_implications(self, triples):
        # Test using manually crafted values
        def imp_func(var, val1, val2):
            if var == "x":
                if (val1, val2) == (1, 2):
                    return 0.85
                elif (val1, val2) == (2, 1):
                    return -0.5
                elif (val1, val2) == (1, 1):  # pragma: no cover
                    # Note: this value should not be used
                    return 10000000  # pragma: no cover
            elif var == "y":
                if (val1, val2) == (2, 3):
                    return 1
                elif (val1, val2) == (3, 2):
                    return 0.0001
            elif var == "z":
                if (val1, val2) == (3, 4):
                    return -0.3
                elif (val1, val2) == (4, 3):
                    return None
            elif var == "w":
                if (val1, val2) == (4, 5):
                    return 0.7
                elif (val1, val2) == (5, 4):
                    return -0.654

        data = Dataset(triples, implication_function=imp_func)
        # Claims are:
        # 0: x=1
        # 1: y=2
        # 2: z=3
        # 3: w=4
        # 4: x=2
        # 5: z=4
        # 6: w=5
        # 7: y=3
        expected_imp = np.array([
            [0, 0, 0, 0, 0.85, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1],
            [0, 0, 0, 0, 0, -0.3, 0, 0],
            [0, 0, 0, 0, 0, 0, 0.7, 0],
            [-0.5, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, -0.654, 0, 0, 0, 0],
            [0, 0.0001, 0, 0, 0, 0, 0, 0]
        ])
        assert data.imp.shape == (8, 8)
        assert np.array_equal(data.imp.toarray(), expected_imp)

    def test_invalid_implication_values(self, triples):
        def too_big(var, val1, val2):
            return 1.001

        def too_small(var, val1, val2):
            return -1.001

        with pytest.raises(ValueError):
            Dataset(triples, implication_function=too_big)
        with pytest.raises(ValueError):
            Dataset(triples, implication_function=too_small)


class TestFileDataset:
    @pytest.fixture
    def example_cls(self):
        class ExampleDataset(FileDataset):
            def get_tuples(self, fileobj):
                for line in map(str.strip, fileobj):
                    source, var, value = line.split(" | ")
                    yield (
                        "source {}".format(source),
                        "var {}".format(var),
                        int(value) * 2 + 1
                    )
        return ExampleDataset

    @pytest.fixture
    def file_contents(self):
        return "\n".join([
            "abc | xyz | 42",
            "def | xyz | 3",
            "abc | XYZ | 7",
            "ghi | XYZ | 7",
            "def | XYZ | 6",
        ])

    def test_base(self, file_contents, tmpdir):
        input_file = tmpdir.join("test_input.dataset")
        input_file.write(file_contents)
        with pytest.raises(NotImplementedError):
            FileDataset(str(input_file))

    def test_basic(self, example_cls, file_contents, tmpdir):
        input_file = tmpdir.join("test_input.dataset")
        input_file.write(file_contents)
        dataset = example_cls(str(input_file))
        # Claims should be:
        # 0: xyz = 85
        # 1: xyz = 7
        # 2: XYZ = 15
        # 3: XYZ = 13
        expected_sc = np.array([
            [1, 0, 1, 0],
            [0, 1, 0, 1],
            [0, 0, 1, 0]
        ])
        assert np.array_equal(dataset.sc.toarray(), expected_sc)

        # Use voting algorithm to get results, and check they are as expected
        res = MajorityVoting().run(dataset)

        assert res.trust == {"source abc": 1, "source def": 1, "source ghi": 1}
        assert res.belief == {
            "var xyz": {85: 0.5, 7: 0.5},
            "var XYZ": {15: 1, 13: 0.5}
        }

    def test_implications(self, example_cls, file_contents, tmpdir):
        """
        Check that claim implications can still be used with file datasets
        """
        def imp(var, val1, val2):
            if var == "var xyz":
                if (val1, val2) == (7, 85):
                    return 1
                else:
                    return -1
            else:
                if (val1, val2) == (13, 15):
                    return 0.5
                else:
                    return -0.5

        input_file = tmpdir.join("test_input.dataset")
        input_file.write(file_contents)
        dataset = example_cls(str(input_file), implication_function=imp)

        expected_imp = np.array([
            [0, -1, 0, 0],
            [1, 0, 0, 0],
            [0, 0, 0, -0.5],
            [0, 0, 0.5, 0]
        ])
        assert np.array_equal(dataset.imp.toarray(), expected_imp)


class TestFileSupervisedData:
    @pytest.fixture
    def example_cls(self):
        class ExampleSupervisedData(FileSupervisedData):
            def get_pairs(self, fileobj):
                for line in map(str.strip, fileobj):
                    var, true_val = line.split(" | ")
                    yield (var, true_val)
        return ExampleSupervisedData

    @pytest.fixture
    def file_contents(self):
        return "\n".join([
            "speed of light | c",
            "best programming language | python",
            "favourite cheese | cheddar"
        ])

    def test_base(self, file_contents, tmpdir):
        input_file = tmpdir.join("sup.dataset")
        input_file.write(file_contents)
        with pytest.raises(NotImplementedError):
            FileSupervisedData(None, str(input_file))

    def test_basic(self, example_cls, file_contents, tmpdir):
        input_file = tmpdir.join("sup.dataset")
        input_file.write(file_contents)
        sup = example_cls(None, str(input_file))
        assert sup.values == {
            "speed of light": "c",
            "best programming language": "python",
            "favourite cheese": "cheddar"
        }



from truthdiscovery.input.dataset import Dataset, IDMapping
from truthdiscovery.input.file_helpers import FileDataset, FileSupervisedData
from truthdiscovery.input.matrix_dataset import MatrixDataset
from truthdiscovery.input.supervised_data import SupervisedData
from truthdiscovery.input.synthetic_data import SyntheticData



import os

import pycodestyle

import truthdiscovery


class TestCodeStyle:
    """
    Check source code style according to pycodestyle.

    The code should always meet the PEP8 guidelines. More specific code style
    checks are done with pylint, but this is not automated since we allow some
    pylint failures
    """
    REPO_ROOT = os.path.dirname(truthdiscovery.__file__)

    def test_pep8(self):
        style = pycodestyle.StyleGuide()
        report = style.check_files(paths=self.get_source_files(self.REPO_ROOT))
        num_errors = report.get_count()
        assert num_errors == 0, "{} pycodestyle error(s)".format(num_errors)

    def get_source_files(self, root_dir):
        for dirpath, dirnames, filenames in os.walk(root_dir):
            for filename in filenames:
                full_path = os.path.join(dirpath, filename)
                if full_path.endswith(".py"):
                    yield full_path



import numpy as np

from truthdiscovery.algorithm.base import BaseIterativeAlgorithm


class Sums(BaseIterativeAlgorithm):
    """
    Sums, or Hubs and Authorities, algorithm.

    Described by Kleinberg for web pages, and adapted to truth discovery by
    Pasternack and Roth
    """
    def _run(self, data):
        trust = np.zeros((data.num_sources,))
        belief = self.get_prior_beliefs(data)
        self.log(data, trust, belief)

        while not self.iterator.finished():
            new_trust = data.sc @ belief
            belief = data.sc.T @ new_trust

            # Trust and belief are normalised so that the largest entries in
            # each are 1; otherwise trust and belief scores grow without bound
            new_trust = new_trust / max(new_trust)
            belief = belief / max(belief)

            self.iterator.compare(trust, new_trust)
            trust = new_trust
            self.log(data, trust, belief)

        return trust, belief



import numpy as np
import numpy.ma as ma

from truthdiscovery.input.matrix_dataset import MatrixDataset
from truthdiscovery.input.supervised_data import SupervisedData


class SyntheticData(SupervisedData):
    """
    A synthetic dataset generated randomly according to given source trust
    values, each of which is interpreted as the probability that a source's
    claim is correct
    """
    def __init__(self, trust, num_variables=100, claim_probability=0.5,
                 domain_size=4, **kwargs):
        """
        :param trust: list or numpy array of trust values in [0, 1] for sources
        :param num_variables: the number of artificial variables to generate
        :param claim_probability: the probability of a source making a claim
                                  about the value of a given variable
        :param domain_size: the number of possible values each variable
                            may take. The possible values are
                            ``[0, .... d - 1]``.
        :raises ValueError: if invalid parameters are given
        """
        if isinstance(trust, list):
            trust = np.array(trust)
        if trust.ndim != 1:
            raise ValueError("Trust vector must be one dimensional")
        if trust.shape[0] == 0:
            raise ValueError("No trust values provided")
        # Check that trust values are numbers in in [0, 1]
        if np.any(np.isnan(trust)) or np.any(trust < 0) or np.any(trust > 1):
            raise ValueError("Trust values must be in [0, 1]")
        if claim_probability <= 0 or claim_probability > 1:
            raise ValueError("Claim probability must be in (0, 1]")
        if domain_size <= 1:
            raise ValueError("Domain size must be greater than 1")

        # Generate 'true' values for the variables uniformly from [0,...,d - 1]
        true_values = np.random.randint(0, domain_size, size=(num_variables,))

        sv_mat = ma.masked_all((len(trust), num_variables))
        for var, true_value in enumerate(true_values):
            claim_made = False
            for source, trust_val in enumerate(trust):
                if np.random.random_sample() <= claim_probability:
                    claim_made = True
                    sv_mat[source, var] = self.generate_claim(
                        trust_val, true_value, domain_size
                    )
            # Make sure at least one source makes a claim about this variable
            if not claim_made:  # pragma: no cover
                source = np.random.randint(0, len(trust))
                sv_mat[source, var] = self.generate_claim(
                    trust[source], true_value, domain_size
                )

        # Make sure all sources make at least one claim
        for source, row in enumerate(sv_mat):
            if row.mask.all():  # pragma: no cover
                var = np.random.randint(0, num_variables)
                sv_mat[source, var] = self.generate_claim(
                    trust[source], true_values[var], domain_size
                )

        # Convert to dict, as required by parent class
        true_values_dict = {var: val for var, val in enumerate(true_values)
                            if not ma.is_masked(val)}
        super().__init__(MatrixDataset(sv_mat), true_values_dict, **kwargs)

    @classmethod
    def generate_claim(cls, trust_val, true_value, domain_size):
        """
        Generate a value for a source to claim for a variable

        :param trust_val: trust value for the source in [0, 1]
        :param true_value: the true value for the variable
        :param domain_size: the number of possible values for the variable; the
                            domain is ``{0, 1, ..., domain_size - 1}``
        """
        # Source claims the correct value with probability trust_val, and
        # chooses an incorrect value uniformly otherwise
        wrong_prob = (1 - trust_val) / (domain_size - 1)
        prob_dist = [wrong_prob] * domain_size
        prob_dist[int(true_value)] = trust_val
        # Draw claimed value from domain with above probability distribution
        return np.random.choice(range(domain_size), p=prob_dist)

    def to_csv(self):
        """
        :return: a string representation of data and generated true values in
                 CSV format
        """
        rows, cols = self.data.sv.shape
        temp = ma.masked_all((rows + 1, cols))
        for var, val in self.values.items():
            # var labels coincide with index in matrix here
            temp[0, var] = val
        temp[1:, :] = self.data.sv
        return MatrixDataset(temp).to_csv()



import copy

import numpy as np

from truthdiscovery.utils import filter_dict


class Result:
    """
    Object to hold the results of truth discovery.
    """
    def __init__(self, trust, belief, time_taken, iterations=None):
        """
        :param trust:  a mapping of the form ``{source_label: trust_val, ..}``
                       containing trust values for sources
        :param belief: a mapping of the form
                       ``{var_label: {val: belief, ...}, ...}`` containing
                       belief values for claims
        :param time_taken: seconds taken to produce these results
        :param iterations: number of iterations the algorithm ran for, or None
                           if not applicable
        """
        self.trust = trust
        self.belief = belief
        self.time_taken = time_taken
        self.iterations = iterations

    def get_most_believed_values(self, var):
        """
        Compute the most believed values for a variable. Note that more than
        one value may be most-believable if there are ties for maximum belief.

        :param var: label of the variable to consider
        :return: a generator of the values with maximum belief
        """
        sorted_beliefs = sorted(self.belief[var].items(), key=lambda x: -x[1])
        max_belief = sorted_beliefs[0][1]

        for val, belief in sorted_beliefs:
            if belief < max_belief:
                break
            yield val

    def filter(self, sources=None, variables=None):
        """
        Filter a set of results to only include trust and belief scores for
        subsets of the sources and variables

        :param sources:   iterable of source labels to narrow the trust scores
                          down to, or None to perform no source filtering
        :param variables: iterable of variable labels to narrow the belief
                          scores down to, or None to perform no variable
                          filtering
        :return: a :any:`Result` object with trust/belief scores for only the
                 requested sources and variables
        """
        new_scores = []
        filters = ((sources, self.trust), (variables, self.belief))
        for filter_set, full_scores in filters:
            if filter_set is not None:
                new_scores.append(dict(filter_dict(full_scores, filter_set)))
            else:
                new_scores.append(copy.deepcopy(full_scores))
        new_trust, new_belief = new_scores

        return Result(new_trust, new_belief, self.time_taken, self.iterations)

    def _get_stats(self, scores_dict):
        """
        :return: ``(mean, stddev)`` of values in ``scores_dict``
        """
        scores_vec = np.fromiter(
            scores_dict.values(), dtype=np.float64, count=len(scores_dict)
        )
        return (np.mean(scores_vec), np.std(scores_vec))

    def get_trust_stats(self):
        """
        :return: a tuple ``(mean, stddev)`` of the mean and standard deviation
                 of trust scores
        """
        return self._get_stats(self.trust)

    def get_belief_stats(self):
        """
        :return: a dictionary of the form ``{var_label: (mean, stddev), ...}``
                 containing mean and standard deviation of belief scores for
                 each variable
        """
        return {
            var: self._get_stats(scores) for var, scores in self.belief.items()
        }
